"""
NFCTSocket -- low level connection tracking API

See also: pyroute2.conntrack
"""

import socket

from pyroute2.netlink import (
    NETLINK_NETFILTER,
    NLA_F_NESTED,
    NLM_F_ACK,
    NLM_F_CREATE,
    NLM_F_DUMP,
    NLM_F_EXCL,
    NLM_F_REQUEST,
    NLMSG_ERROR,
    nla,
)
from pyroute2.netlink.core import SyncAPI
from pyroute2.netlink.nfnetlink import NFNL_SUBSYS_CTNETLINK, nfgen_msg
from pyroute2.netlink.nlsocket import AsyncNetlinkSocket

IPCTNL_MSG_CT_NEW = 0
IPCTNL_MSG_CT_GET = 1
IPCTNL_MSG_CT_DELETE = 2
IPCTNL_MSG_CT_GET_CTRZERO = 3
IPCTNL_MSG_CT_GET_STATS_CPU = 4
IPCTNL_MSG_CT_GET_STATS = 5
IPCTNL_MSG_CT_GET_DYING = 6
IPCTNL_MSG_CT_GET_UNCONFIRMED = 7
IPCTNL_MSG_MAX = 8


try:
    IP_PROTOCOLS = {
        num: name[8:]
        for name, num in vars(socket).items()
        if name.startswith("IPPROTO")
    }
except (IOError, OSError):
    IP_PROTOCOLS = {}


# Window scaling is advertised by the sender
IP_CT_TCP_FLAG_WINDOW_SCALE = 0x01

# SACK is permitted by the sender
IP_CT_TCP_FLAG_SACK_PERM = 0x02

# This sender sent FIN first
IP_CT_TCP_FLAG_CLOSE_INIT = 0x04

# Be liberal in window checking
IP_CT_TCP_FLAG_BE_LIBERAL = 0x08

# Has unacknowledged data
IP_CT_TCP_FLAG_DATA_UNACKNOWLEDGED = 0x10

# The field td_maxack has been set
IP_CT_TCP_FLAG_MAXACK_SET = 0x20


# From linux/include/net/tcp_states.h
TCPF_ESTABLISHED = 1 << 1
TCPF_SYN_SENT = 1 << 2
TCPF_SYN_RECV = 1 << 3
TCPF_FIN_WAIT1 = 1 << 4
TCPF_FIN_WAIT2 = 1 << 5
TCPF_TIME_WAIT = 1 << 6
TCPF_CLOSE = 1 << 7
TCPF_CLOSE_WAIT = 1 << 8
TCPF_LAST_ACK = 1 << 9
TCPF_LISTEN = 1 << 10
TCPF_CLOSING = 1 << 11
TCPF_NEW_SYN_RECV = 1 << 12
TCPF_TO_NAME = {
    TCPF_ESTABLISHED: 'ESTABLISHED',
    TCPF_SYN_SENT: 'SYN_SENT',
    TCPF_SYN_RECV: 'SYN_RECV',
    TCPF_FIN_WAIT1: 'FIN_WAIT1',
    TCPF_FIN_WAIT2: 'FIN_WAIT2',
    TCPF_TIME_WAIT: 'TIME_WAIT',
    TCPF_CLOSE: 'CLOSE',
    TCPF_CLOSE_WAIT: 'CLOSE_WAIT',
    TCPF_LAST_ACK: 'LAST_ACK',
    TCPF_LISTEN: 'LISTEN',
    TCPF_CLOSING: 'CLOSING',
    TCPF_NEW_SYN_RECV: 'NEW_SYN_RECV',
}


# From include/uapi/linux/netfilter/nf_conntrack_common.h
IPS_EXPECTED = 1 << 0
IPS_SEEN_REPLY = 1 << 1
IPS_ASSURED = 1 << 2
IPS_CONFIRMED = 1 << 3
IPS_SRC_NAT = 1 << 4
IPS_DST_NAT = 1 << 5
IPS_NAT_MASK = IPS_DST_NAT | IPS_SRC_NAT
IPS_SEQ_ADJUST = 1 << 6
IPS_SRC_NAT_DONE = 1 << 7
IPS_DST_NAT_DONE = 1 << 8
IPS_NAT_DONE_MASK = IPS_DST_NAT_DONE | IPS_SRC_NAT_DONE
IPS_DYING = 1 << 9
IPS_FIXED_TIMEOUT = 1 << 10
IPS_TEMPLATE = 1 << 11
IPS_UNTRACKED = 1 << 12
IPS_HELPER = 1 << 13
IPS_OFFLOAD = 1 << 14
IPS_UNCHANGEABLE_MASK = (
    IPS_NAT_DONE_MASK
    | IPS_NAT_MASK
    | IPS_EXPECTED
    | IPS_CONFIRMED
    | IPS_DYING
    | IPS_SEQ_ADJUST
    | IPS_TEMPLATE
    | IPS_OFFLOAD
)
IPSBIT_TO_NAME = {
    IPS_EXPECTED: 'EXPECTED',
    IPS_SEEN_REPLY: 'SEEN_REPLY',
    IPS_ASSURED: 'ASSURED',
    IPS_CONFIRMED: 'CONFIRMED',
    IPS_SRC_NAT: 'SRC_NAT',
    IPS_DST_NAT: 'DST_NAT',
    IPS_SEQ_ADJUST: 'SEQ_ADJUST',
    IPS_SRC_NAT_DONE: 'SRC_NAT_DONE',
    IPS_DST_NAT_DONE: 'DST_NAT_DONE',
    IPS_DYING: 'DYING',
    IPS_FIXED_TIMEOUT: 'FIXED_TIMEOUT',
    IPS_TEMPLATE: 'TEMPLATE',
    IPS_UNTRACKED: 'UNTRACKED',
    IPS_HELPER: 'HELPER',
    IPS_OFFLOAD: 'OFFLOAD',
}

# From include/uapi/linux/netfilter/nf_conntrack_tcp.h
IP_CT_TCP_FLAG_WINDOW_SCALE = 0x01
IP_CT_TCP_FLAG_SACK_PERM = 0x02
IP_CT_TCP_FLAG_CLOSE_INIT = 0x04
IP_CT_TCP_FLAG_BE_LIBERAL = 0x08
IP_CT_TCP_FLAG_DATA_UNACKNOWLEDGED = 0x10
IP_CT_TCP_FLAG_MAXACK_SET = 0x20
IP_CT_EXP_CHALLENGE_ACK = 0x40
IP_CT_TCP_SIMULTANEOUS_OPEN = 0x80
IP_CT_TCP_FLAG_TO_NAME = {
    IP_CT_TCP_FLAG_WINDOW_SCALE: 'WINDOW_SCALE',
    IP_CT_TCP_FLAG_SACK_PERM: 'SACK_PERM',
    IP_CT_TCP_FLAG_CLOSE_INIT: 'CLOSE_INIT',
    IP_CT_TCP_FLAG_BE_LIBERAL: 'BE_LIBERAL',
    IP_CT_TCP_FLAG_DATA_UNACKNOWLEDGED: 'DATA_UNACKNOWLEDGED',
    IP_CT_TCP_FLAG_MAXACK_SET: 'MAXACK_SET',
    IP_CT_EXP_CHALLENGE_ACK: 'CHALLENGE_ACK',
    IP_CT_TCP_SIMULTANEOUS_OPEN: 'SIMULTANEOUS_OPEN',
}

# From linux/include/uapi/linux/netfilter/nf_conntrack_tcp.h
TCP_CONNTRACK_SYN_SENT = 1
TCP_CONNTRACK_SYN_RECV = 2
TCP_CONNTRACK_ESTABLISHED = 3
TCP_CONNTRACK_FIN_WAIT = 4
TCP_CONNTRACK_CLOSE_WAIT = 5
TCP_CONNTRACK_LAST_ACK = 6
TCP_CONNTRACK_TIME_WAIT = 7
TCP_CONNTRACK_CLOSE = 8
TCP_CONNTRACK_LISTEN = 9
TCP_CONNTRACK_MAX = 10
TCP_CONNTRACK_IGNORE = 11
TCP_CONNTRACK_RETRANS = 12
TCP_CONNTRACK_UNACK = 13
TCP_CONNTRACK_TIMEOUT_MAX = 14
TCP_CONNTRACK_TO_NAME = {
    TCP_CONNTRACK_SYN_SENT: "SYN_SENT",
    TCP_CONNTRACK_SYN_RECV: "SYN_RECV",
    TCP_CONNTRACK_ESTABLISHED: "ESTABLISHED",
    TCP_CONNTRACK_FIN_WAIT: "FIN_WAIT",
    TCP_CONNTRACK_CLOSE_WAIT: "CLOSE_WAIT",
    TCP_CONNTRACK_LAST_ACK: "LAST_ACK",
    TCP_CONNTRACK_TIME_WAIT: "TIME_WAIT",
    TCP_CONNTRACK_CLOSE: "CLOSE",
    TCP_CONNTRACK_LISTEN: "LISTEN",
    TCP_CONNTRACK_MAX: "MAX",
    TCP_CONNTRACK_IGNORE: "IGNORE",
    TCP_CONNTRACK_RETRANS: "RETRANS",
    TCP_CONNTRACK_UNACK: "UNACK",
    TCP_CONNTRACK_TIMEOUT_MAX: "TIMEOUT_MAX",
}


def terminate_single_msg(msg):
    return msg


def terminate_error_msg(msg):
    return msg['header']['type'] == NLMSG_ERROR


class nfct_stats(nfgen_msg):
    prefix = 'CTA_STATS_'
    nla_map = (
        ('CTA_STATS_GLOBAL_UNSPEC', 'none'),
        ('CTA_STATS_GLOBAL_ENTRIES', 'be32'),
        ('CTA_STATS_GLOBAL_MAX_ENTRIES', 'be32'),
    )


class nfct_stats_cpu(nfgen_msg):
    prefix = 'CTA_STATS_'
    nla_map = (
        ('CTA_STATS_UNSPEC', 'none'),
        ('CTA_STATS_SEARCHED', 'be32'),
        ('CTA_STATS_FOUND', 'be32'),
        ('CTA_STATS_NEW', 'be32'),
        ('CTA_STATS_INVALID', 'be32'),
        ('CTA_STATS_IGNORE', 'be32'),
        ('CTA_STATS_DELETE', 'be32'),
        ('CTA_STATS_DELETE_LIST', 'be32'),
        ('CTA_STATS_INSERT', 'be32'),
        ('CTA_STATS_INSERT_FAILED', 'be32'),
        ('CTA_STATS_DROP', 'be32'),
        ('CTA_STATS_EARLY_DROP', 'be32'),
        ('CTA_STATS_ERROR', 'be32'),
        ('CTA_STATS_SEARCH_RESTART', 'be32'),
    )


class nfct_msg(nfgen_msg):
    prefix = 'CTA_'
    nla_map = (
        ('CTA_UNSPEC', 'none'),
        ('CTA_TUPLE_ORIG', 'cta_tuple'),
        ('CTA_TUPLE_REPLY', 'cta_tuple'),
        ('CTA_STATUS', 'be32'),
        ('CTA_PROTOINFO', 'cta_protoinfo'),
        ('CTA_HELP', 'asciiz'),
        ('CTA_NAT_SRC', 'cta_nat'),
        ('CTA_TIMEOUT', 'be32'),
        ('CTA_MARK', 'be32'),
        ('CTA_COUNTERS_ORIG', 'cta_counters'),
        ('CTA_COUNTERS_REPLY', 'cta_counters'),
        ('CTA_USE', 'be32'),
        ('CTA_ID', 'be32'),
        ('CTA_NAT_DST', 'cta_nat'),
        ('CTA_TUPLE_MASTER', 'cta_tuple'),
        ('CTA_SEQ_ADJ_ORIG', 'cta_nat_seq_adj'),
        ('CTA_SEQ_ADJ_REPLY', 'cta_nat_seq_adj'),
        ('CTA_SECMARK', 'be32'),
        ('CTA_ZONE', 'be16'),
        ('CTA_SECCTX', 'cta_secctx'),
        ('CTA_TIMESTAMP', 'cta_timestamp'),
        ('CTA_MARK_MASK', 'be32'),
        ('CTA_LABELS', 'cta_labels'),
        ('CTA_LABELS_MASK', 'cta_labels'),
        ('CTA_SYNPROXY', 'cta_synproxy'),
        ('CTA_FILTER', 'cta_filter'),
        ('CTA_STATUS_MASK', 'be32'),
    )

    @classmethod
    def create_from(cls, **kwargs):
        self = cls()

        for key, value in kwargs.items():
            if isinstance(value, NFCTAttr):
                value = {'attrs': value.attrs()}
            if value is not None:
                self['attrs'].append([self.name2nla(key), value])

        return self

    class cta_tuple(nla):
        prefix = 'CTA_TUPLE_'
        nla_map = (
            ('CTA_TUPLE_UNSPEC', 'none'),
            ('CTA_TUPLE_IP', 'cta_ip'),
            ('CTA_TUPLE_PROTO', 'cta_proto'),
        )

        class cta_ip(nla):
            prefix = 'CTA_IP_'
            nla_map = (
                ('CTA_IP_UNSPEC', 'none'),
                ('CTA_IP_V4_SRC', 'ip4addr'),
                ('CTA_IP_V4_DST', 'ip4addr'),
                ('CTA_IP_V6_SRC', 'ip6addr'),
                ('CTA_IP_V6_DST', 'ip6addr'),
            )

        class cta_proto(nla):
            prefix = 'CTA_PROTO_'
            nla_map = (
                ('CTA_PROTO_UNSPEC', 'none'),
                ('CTA_PROTO_NUM', 'uint8'),
                ('CTA_PROTO_SRC_PORT', 'be16'),
                ('CTA_PROTO_DST_PORT', 'be16'),
                ('CTA_PROTO_ICMP_ID', 'be16'),
                ('CTA_PROTO_ICMP_TYPE', 'uint8'),
                ('CTA_PROTO_ICMP_CODE', 'uint8'),
                ('CTA_PROTO_ICMPV6_ID', 'be16'),
                ('CTA_PROTO_ICMPV6_TYPE', 'uint8'),
                ('CTA_PROTO_ICMPV6_CODE', 'uint8'),
            )

    class cta_protoinfo(nla):
        nla_map = (
            ('CTA_PROTOINFO_UNSPEC', 'none'),
            ('CTA_PROTOINFO_TCP', 'cta_protoinfo_tcp'),
            ('CTA_PROTOINFO_DCCP', 'cta_protoinfo_dccp'),
            ('CTA_PROTOINFO_SCTP', 'cta_protoinfo_sctp'),
        )

        class cta_protoinfo_tcp(nla):
            nla_map = (
                ('CTA_PROTOINFO_TCP_UNSPEC', 'none'),
                ('CTA_PROTOINFO_TCP_STATE', 'uint8'),
                ('CTA_PROTOINFO_TCP_WSCALE_ORIGINAL', 'uint8'),
                ('CTA_PROTOINFO_TCP_WSCALE_REPLY', 'uint8'),
                ('CTA_PROTOINFO_TCP_FLAGS_ORIGINAL', 'cta_tcp_flags'),
                ('CTA_PROTOINFO_TCP_FLAGS_REPLY', 'cta_tcp_flags'),
            )

            class cta_tcp_flags(nla):
                fields = [('value', 'BB')]

        class cta_protoinfo_dccp(nla):
            nla_map = (
                ('CTA_PROTOINFO_DCCP_UNSPEC', 'none'),
                ('CTA_PROTOINFO_DCCP_STATE', 'uint8'),
                ('CTA_PROTOINFO_DCCP_ROLE', 'uint8'),
                ('CTA_PROTOINFO_DCCP_HANDSHAKE_SEQ', 'be64'),
            )

        class cta_protoinfo_sctp(nla):
            nla_map = (
                ('CTA_PROTOINFO_SCTP_UNSPEC', 'none'),
                ('CTA_PROTOINFO_SCTP_STATE', 'uint8'),
                ('CTA_PROTOINFO_SCTP_VTAG_ORIGINAL', 'be32'),
                ('CTA_PROTOINFO_SCTP_VTAG_REPLY', 'be32'),
            )

    class cta_nat(nla):
        nla_map = (
            ('CTA_NAT_UNSPEC', 'none'),
            ('CTA_NAT_V4_MINIP', 'ip4addr'),
            ('CTA_NAT_V4_MAXIP', 'ip4addr'),
            ('CTA_NAT_PROTO', 'cta_protonat'),
            ('CTA_NAT_V6_MINIP', 'ip6addr'),
            ('CTA_NAT_V6_MAXIP', 'ip6addr'),
        )

        class cta_protonat(nla):
            nla_map = (
                ('CTA_PROTONAT_UNSPEC', 'none'),
                ('CTA_PROTONAT_PORT_MIN', 'be16'),
                ('CTA_PROTONAT_PORT_MAX', 'be16'),
            )

    class cta_nat_seq_adj(nla):
        nla_map = (
            ('CTA_NAT_SEQ_UNSPEC', 'none'),
            ('CTA_NAT_SEQ_CORRECTION_POS', 'be32'),
            ('CTA_NAT_SEQ_OFFSET_BEFORE', 'be32'),
            ('CTA_NAT_SEQ_OFFSET_AFTER', 'be32'),
        )

    class cta_counters(nla):
        nla_map = (
            ('CTA_COUNTERS_UNSPEC', 'none'),
            ('CTA_COUNTERS_PACKETS', 'be64'),
            ('CTA_COUNTERS_BYTES', 'be64'),
            ('CTA_COUNTERS32_PACKETS', 'be32'),
            ('CTA_COUNTERS32_BYTES', 'be32'),
        )

    class cta_secctx(nla):
        nla_map = (
            ('CTA_SECCTX_UNSPEC', 'none'),
            ('CTA_SECCTX_NAME', 'asciiz'),
        )

    class cta_timestamp(nla):
        nla_map = (
            ('CTA_TIMESTAMP_UNSPEC', 'none'),
            ('CTA_TIMESTAMP_START', 'be64'),
            ('CTA_TIMESTAMP_STOP', 'be64'),
        )

    class cta_filter(nla):
        nla_flags = NLA_F_NESTED
        nla_map = (
            ('CTA_FILTER_UNSPEC', 'none'),
            ('CTA_FILTER_ORIG_FLAGS', 'uint32'),
            ('CTA_FILTER_REPLY_FLAGS', 'uint32'),
        )

    class cta_labels(nla):
        fields = [('value', 'QQ')]

        def encode(self):
            if not isinstance(self['value'], tuple):
                self['value'] = (
                    self['value'] & 0xFFFFFFFFFFFFFFFF,
                    self['value'] >> 64,
                )
            nla.encode(self)

        def decode(self):
            nla.decode(self)
            if isinstance(self['value'], tuple):
                self['value'] = (self['value'][0] & 0xFFFFFFFFFFFFFFFF) | (
                    self['value'][1] << 64
                )

    class cta_synproxy(nla):
        nla_map = (
            ('CTA_SYNPROXY_UNSPEC', 'none'),
            ('CTA_SYNPROXY_ISN', 'be32'),
            ('CTA_SYNPROXY_ITS', 'be32'),
            ('CTA_SYNPROXY_TSOFF', 'be32'),
        )


FILTER_FLAG_CTA_IP_SRC = 1 << 0
FILTER_FLAG_CTA_IP_DST = 1 << 1
FILTER_FLAG_CTA_TUPLE_ZONE = 1 << 2
FILTER_FLAG_CTA_PROTO_NUM = 1 << 3
FILTER_FLAG_CTA_PROTO_SRC_PORT = 1 << 4
FILTER_FLAG_CTA_PROTO_DST_PORT = 1 << 5
FILTER_FLAG_CTA_PROTO_ICMP_TYPE = 1 << 6
FILTER_FLAG_CTA_PROTO_ICMP_CODE = 1 << 7
FILTER_FLAG_CTA_PROTO_ICMP_ID = 1 << 8
FILTER_FLAG_CTA_PROTO_ICMPV6_TYPE = 1 << 9
FILTER_FLAG_CTA_PROTO_ICMPV6_CODE = 1 << 10
FILTER_FLAG_CTA_PROTO_ICMPV6_ID = 1 << 11

FILTER_FLAG_ALL_CTA_PROTO = (
    FILTER_FLAG_CTA_PROTO_SRC_PORT
    | FILTER_FLAG_CTA_PROTO_DST_PORT
    | FILTER_FLAG_CTA_PROTO_ICMP_TYPE
    | FILTER_FLAG_CTA_PROTO_ICMP_CODE
    | FILTER_FLAG_CTA_PROTO_ICMP_ID
    | FILTER_FLAG_CTA_PROTO_ICMPV6_TYPE
    | FILTER_FLAG_CTA_PROTO_ICMPV6_CODE
    | FILTER_FLAG_CTA_PROTO_ICMPV6_ID
)
FILTER_FLAG_ALL = 0xFFFFFFFF


class NFCTAttr(object):
    def attrs(self):
        return []


class NFCTAttrTuple(NFCTAttr):
    __slots__ = (
        'saddr',
        'daddr',
        'proto',
        'sport',
        'dport',
        'icmp_id',
        'icmp_type',
        'family',
        '_attr_ip',
        '_attr_icmp',
    )

    def __init__(
        self,
        family=socket.AF_INET,
        saddr=None,
        daddr=None,
        proto=None,
        sport=None,
        dport=None,
        icmp_id=None,
        icmp_type=None,
        icmp_code=None,
    ):
        self.saddr = saddr
        self.daddr = daddr
        self.proto = proto
        self.sport = sport
        self.dport = dport
        self.icmp_id = icmp_id
        self.icmp_type = icmp_type
        self.icmp_code = icmp_code
        self.family = family

        self._attr_ip, self._attr_icmp = {
            socket.AF_INET: ['CTA_IP_V4', 'CTA_PROTO_ICMP'],
            socket.AF_INET6: ['CTA_IP_V6', 'CTA_PROTO_ICMPV6'],
        }[self.family]

    def proto_name(self):
        return IP_PROTOCOLS.get(self.proto, None)

    def reverse(self):
        return NFCTAttrTuple(
            family=self.family,
            saddr=self.daddr,
            daddr=self.saddr,
            proto=self.proto,
            sport=self.dport,
            dport=self.sport,
            icmp_id=self.icmp_id,
            icmp_type=self.icmp_type,
            icmp_code=self.icmp_code,
        )

    def attrs(self):
        cta_ip = []
        cta_proto = []
        cta_tuple = []

        self.flags = 0

        if self.saddr is not None:
            cta_ip.append([self._attr_ip + '_SRC', self.saddr])
            self.flags |= FILTER_FLAG_CTA_IP_SRC

        if self.daddr is not None:
            cta_ip.append([self._attr_ip + '_DST', self.daddr])
            self.flags |= FILTER_FLAG_CTA_IP_DST

        if self.proto is not None:
            cta_proto.append(['CTA_PROTO_NUM', self.proto])
            self.flags |= FILTER_FLAG_CTA_PROTO_NUM

        if self.sport is not None:
            cta_proto.append(['CTA_PROTO_SRC_PORT', self.sport])
            self.flags |= FILTER_FLAG_CTA_PROTO_SRC_PORT

        if self.dport is not None:
            cta_proto.append(['CTA_PROTO_DST_PORT', self.dport])
            self.flags |= FILTER_FLAG_CTA_PROTO_DST_PORT

        if self.icmp_id is not None:
            cta_proto.append([self._attr_icmp + '_ID', self.icmp_id])

        if self.icmp_type is not None:
            cta_proto.append([self._attr_icmp + '_TYPE', self.icmp_type])

        if self.icmp_code is not None:
            cta_proto.append([self._attr_icmp + '_CODE', self.icmp_code])

        if cta_ip:
            cta_tuple.append(['CTA_TUPLE_IP', {'attrs': cta_ip}])

        if cta_proto:
            cta_tuple.append(['CTA_TUPLE_PROTO', {'attrs': cta_proto}])

        return cta_tuple

    @classmethod
    def from_netlink(cls, family, ndmsg):
        cta_ip = ndmsg.get_attr('CTA_TUPLE_IP')
        cta_proto = ndmsg.get_attr('CTA_TUPLE_PROTO')
        kwargs = {'family': family}

        if family == socket.AF_INET:
            kwargs['saddr'] = cta_ip.get_attr('CTA_IP_V4_SRC')
            kwargs['daddr'] = cta_ip.get_attr('CTA_IP_V4_DST')
        elif family == socket.AF_INET6:
            kwargs['saddr'] = cta_ip.get_attr('CTA_IP_V6_SRC')
            kwargs['daddr'] = cta_ip.get_attr('CTA_IP_V6_DST')
        else:
            raise NotImplementedError(family)

        proto = cta_proto.get_attr('CTA_PROTO_NUM')
        kwargs['proto'] = proto

        if proto == socket.IPPROTO_ICMP:
            kwargs['icmp_id'] = cta_proto.get_attr('CTA_PROTO_ICMP_ID')
            kwargs['icmp_type'] = cta_proto.get_attr('CTA_PROTO_ICMP_TYPE')
            kwargs['icmp_code'] = cta_proto.get_attr('CTA_PROTO_ICMP_CODE')
        elif proto == socket.IPPROTO_ICMPV6:
            kwargs['icmp_id'] = cta_proto.get_attr('CTA_PROTO_ICMPV6_ID')
            kwargs['icmp_type'] = cta_proto.get_attr('CTA_PROTO_ICMPV6_TYPE')
            kwargs['icmp_code'] = cta_proto.get_attr('CTA_PROTO_ICMPV6_CODE')
        elif proto in (socket.IPPROTO_TCP, socket.IPPROTO_UDP):
            kwargs['sport'] = cta_proto.get_attr('CTA_PROTO_SRC_PORT')
            kwargs['dport'] = cta_proto.get_attr('CTA_PROTO_DST_PORT')

        return cls(**kwargs)

    def is_attr_match(self, other, attrname):
        l_attr = getattr(self, attrname)
        if l_attr is not None:
            r_attr = getattr(other, attrname)
            if l_attr != r_attr:
                return False
        return True

    def nla_eq(self, family, ndmsg):
        if self.family != family:
            return False

        test_attr = []
        cta_ip = ndmsg.get_attr('CTA_TUPLE_IP')
        if family == socket.AF_INET:
            test_attr.append((self.saddr, cta_ip, 'CTA_IP_V4_SRC'))
            test_attr.append((self.daddr, cta_ip, 'CTA_IP_V4_DST'))
        elif family == socket.AF_INET6:
            test_attr.append((self.saddr, cta_ip, 'CTA_IP_V6_SRC'))
            test_attr.append((self.daddr, cta_ip, 'CTA_IP_V6_DST'))
        else:
            raise NotImplementedError(family)

        if self.proto is not None:
            cta_proto = ndmsg.get_attr('CTA_TUPLE_PROTO')
            if self.proto != cta_proto.get_attr('CTA_PROTO_NUM'):
                return False

            if self.proto == socket.IPPROTO_ICMP:
                (
                    test_attr.append(
                        (self.icmp_id, cta_proto, 'CTA_PROTO_ICMP_ID')
                    )
                )
                (
                    test_attr.append(
                        (self.icmp_type, cta_proto, 'CTA_PROTO_ICMP_TYPE')
                    )
                )
                (
                    test_attr.append(
                        (self.icmp_code, cta_proto, 'CTA_PROTO_ICMP_CODE')
                    )
                )
            elif self.proto == socket.IPPROTO_ICMPV6:
                (
                    test_attr.append(
                        (self.icmp_id, cta_proto, 'CTA_PROTO_ICMPV6_ID')
                    )
                )
                (
                    test_attr.append(
                        (self.icmp_type, cta_proto, 'CTA_PROTO_ICMPV6_TYPE')
                    )
                )
                (
                    test_attr.append(
                        (self.icmp_code, cta_proto, 'CTA_PROTO_ICMPV6_CODE')
                    )
                )
            elif self.proto in (socket.IPPROTO_TCP, socket.IPPROTO_UDP):
                (
                    test_attr.append(
                        (self.sport, cta_proto, 'CTA_PROTO_SRC_PORT')
                    )
                )
                (
                    test_attr.append(
                        (self.dport, cta_proto, 'CTA_PROTO_DST_PORT')
                    )
                )

        for val, ndmsg, attrname in test_attr:
            if val is not None and val != ndmsg.get_attr(attrname):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            raise NotImplementedError()

        if self.family != other.family:
            return False

        for attrname in ('saddr', 'daddr'):
            if not self.is_attr_match(other, attrname):
                return False

        if self.proto is not None:
            if self.proto != other.proto:
                return False

            if self.proto in (socket.IPPROTO_UDP, socket.IPPROTO_TCP):
                for attrname in ('sport', 'dport'):
                    if not self.is_attr_match(other, attrname):
                        return False
            elif self.proto in (socket.IPPROTO_ICMP, socket.IPPROTO_ICMPV6):
                for attrname in ('icmp_id', 'icmp_type', 'icmp_code'):
                    if not self.is_attr_match(other, attrname):
                        return False

        return True

    def __repr__(self):
        proto_name = self.proto_name()
        if proto_name is None:
            proto_name = 'UNKNOWN'

        if self.family == socket.AF_INET:
            r = 'IPv4('
        elif self.family == socket.AF_INET6:
            r = 'IPv6('
        else:
            r = 'Unkown[family={}]('.format(self.family)
        r += 'saddr={}, daddr={}, '.format(self.saddr, self.daddr)

        r += '{}('.format(proto_name)
        if self.proto in (socket.IPPROTO_ICMP, socket.IPPROTO_ICMPV6):
            r += 'id={}, type={}, code={}'.format(
                self.icmp_id, self.icmp_type, self.icmp_code
            )
        elif self.proto in (socket.IPPROTO_TCP, socket.IPPROTO_UDP):
            r += 'sport={}, dport={}'.format(self.sport, self.dport)
        return r + '))'


class AsyncNFCTSocket(AsyncNetlinkSocket):
    policy = {
        k | (NFNL_SUBSYS_CTNETLINK << 8): v
        for k, v in {
            IPCTNL_MSG_CT_NEW: nfct_msg,
            IPCTNL_MSG_CT_GET: nfct_msg,
            IPCTNL_MSG_CT_DELETE: nfct_msg,
            IPCTNL_MSG_CT_GET_CTRZERO: nfct_msg,
            IPCTNL_MSG_CT_GET_STATS_CPU: nfct_stats_cpu,
            IPCTNL_MSG_CT_GET_STATS: nfct_stats,
            IPCTNL_MSG_CT_GET_DYING: nfct_msg,
            IPCTNL_MSG_CT_GET_UNCONFIRMED: nfct_msg,
        }.items()
    }

    def __init__(self, nfgen_family=socket.AF_INET, **kwargs):
        super().__init__(family=NETLINK_NETFILTER, **kwargs)
        self.register_policy(self.policy)
        self._nfgen_family = nfgen_family

    async def request(self, msg, msg_type, **kwargs):
        msg['nfgen_family'] = self._nfgen_family
        msg_type |= NFNL_SUBSYS_CTNETLINK << 8
        return await self.nlm_request(msg, msg_type, **kwargs)

    async def dump(
        self,
        mark=None,
        mark_mask=0xFFFFFFFF,
        tuple_orig=None,
        tuple_reply=None,
        status=None,
        status_mask=None,
    ):
        """Dump conntrack entries

        Several kernel side filtering are supported:
          * mark and mark_mask, for almost all kernel
          * tuple_orig and tuple_reply, since kernel 5.8 and newer.
            Warning: tuple_reply has a bug in kernel, fixed only recently.
          * status and status_mask, available since kernel 5.15

        tuple_orig and tuple_reply are type NFCTAttrTuple.
        You can give only some attribute for filtering.

        Example::
            # Get only connections from 192.168.1.1
            filter = NFCTAttrTuple(saddr='192.168.1.1')
            ct.dump_entries(tuple_orig=filter)

            # Get HTTPS connections
            filter = NFCTAttrTuple(proto=socket.IPPROTO_TCP, dport=443)
            ct.dump_entries(tuple_orig=filter)

            # Get only one way connection (no reply)
            ct.dump_entries(status=0, status_mask=IPS_SEEN_REPLY)

        Note that NFCTAttrTuple attributes are working like one AND operator.

        Example::
           # Get connections from 192.168.1.1 AND on port 443
           TCP = socket.IPPROTO_TCP
           filter = NFCTAttrTuple(saddr='192.168.1.1', proto=TCP, dport=443)
           ct.dump_entries(tuple_orig=filter)

        """
        if tuple_orig is not None:
            tuple_orig.attrs()  # for creating flags
            cta_filter = {
                'attrs': [['CTA_FILTER_ORIG_FLAGS', tuple_orig.flags]]
            }
            msg = nfct_msg.create_from(
                tuple_orig=tuple_orig, cta_filter=cta_filter
            )
        elif tuple_reply is not None:
            tuple_reply.attrs()
            cta_filter = {
                'attrs': [['CTA_FILTER_REPLY_FLAGS', tuple_reply.flags]]
            }
            msg = nfct_msg.create_from(
                tuple_reply=tuple_reply, cta_filter=cta_filter
            )
        else:
            kwargs = {}
            if mark:
                kwargs['mark'] = mark
                kwargs['mark_mask'] = mark_mask
            if status is not None:
                kwargs['status'] = status
            if status_mask is not None:
                kwargs['status_mask'] = status_mask
            msg = nfct_msg.create_from(**kwargs)
        return await self.request(
            msg, IPCTNL_MSG_CT_GET, msg_flags=NLM_F_REQUEST | NLM_F_DUMP
        )

    async def stat(self):
        return [
            x
            async for x in await self.request(
                nfct_msg(),
                IPCTNL_MSG_CT_GET_STATS_CPU,
                msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
            )
        ]

    async def count(self):
        return [
            x
            async for x in await self.request(
                nfct_msg(),
                IPCTNL_MSG_CT_GET_STATS,
                msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
                terminate=terminate_single_msg,
            )
        ]

    async def flush(self, mark=None, mark_mask=None):
        msg = nfct_msg.create_from(mark=mark, mark_mask=mark_mask)
        return [
            x
            async for x in await self.request(
                msg,
                IPCTNL_MSG_CT_DELETE,
                msg_flags=NLM_F_REQUEST | NLM_F_ACK,
                terminate=terminate_error_msg,
            )
        ]

    async def conntrack_max_size(self):
        return [
            x
            async for x in await self.request(
                nfct_msg(),
                IPCTNL_MSG_CT_GET_STATS,
                msg_flags=NLM_F_REQUEST | NLM_F_DUMP,
                terminate=terminate_single_msg,
            )
        ]

    async def entry(self, cmd, **kwargs):
        """
        Get or change a conntrack entry.

        Examples::
            # add an entry
            ct.entry('add', timeout=30,
                     tuple_orig=NFCTAttrTuple(
                         saddr='192.168.122.1', daddr='192.168.122.67',
                         proto=6, sport=34857, dport=5599),
                     tuple_reply=NFCTAttrTuple(
                         saddr='192.168.122.67', daddr='192.168.122.1',
                         proto=6, sport=5599, dport=34857))

            # set mark=5 on the matching entry
            ct.entry('set', mark=5,
                     tuple_orig=NFCTAttrTuple(
                         saddr='192.168.122.1', daddr='192.168.122.67',
                         proto=6, sport=34857, dport=5599))

            # get an entry
            ct.entry('get',
                     tuple_orig=NFCTAttrTuple(
                         saddr='192.168.122.1', daddr='192.168.122.67',
                         proto=6, sport=34857, dport=5599))

            # delete an entry
            ct.entry('del',
                     tuple_orig=NFCTAttrTuple(
                         saddr='192.168.122.1', daddr='192.168.122.67',
                         proto=6, sport=34857, dport=5599))
        """
        msg_type, msg_flags = {
            'add': [IPCTNL_MSG_CT_NEW, NLM_F_ACK | NLM_F_EXCL | NLM_F_CREATE],
            'set': [IPCTNL_MSG_CT_NEW, NLM_F_ACK],
            'get': [IPCTNL_MSG_CT_GET, NLM_F_ACK],
            'del': [IPCTNL_MSG_CT_DELETE, NLM_F_ACK],
        }[cmd]

        if msg_type == IPCTNL_MSG_CT_DELETE and not (
            'tuple_orig' in kwargs or 'tuple_reply' in kwargs
        ):
            raise ValueError('Deletion requires a tuple at least')

        return [
            x
            async for x in await self.request(
                nfct_msg.create_from(**kwargs),
                msg_type,
                msg_flags=NLM_F_REQUEST | msg_flags,
                terminate=terminate_error_msg,
            )
        ]


class NFCTSocket(SyncAPI):
    def __init__(self, nfgen_family=socket.AF_INET, **kwargs):
        self.asyncore = AsyncNFCTSocket(nfgen_family, **kwargs)

    def request(self, msg, msg_type, **kwargs):
        return self._run_sync_cleanup(
            self.asyncore.request, msg, msg_type, **kwargs
        )

    def dump(
        self,
        mark=None,
        mark_mask=0xFFFFFFFF,
        tuple_orig=None,
        tuple_reply=None,
        status=None,
        status_mask=None,
    ):
        return self._generate_with_cleanup(
            self.asyncore.dump,
            mark,
            mark_mask,
            tuple_orig,
            tuple_reply,
            status,
            status_mask,
        )

    def stat(self):
        return self._run_with_cleanup(self.asyncore.stat)

    def count(self):
        return self._run_with_cleanup(self.asyncore.count)

    def flush(self, mark=None, mark_mask=None):
        return self._run_with_cleanup(self.asyncore.flush, mark, mark_mask)

    def conntrack_max_size(self):
        return self._run_with_cleanup(self.asyncore.conntrack_max_size)

    def entry(self, cmd, **kwarg):
        return self._run_with_cleanup(self.asyncore.entry, cmd, **kwarg)



'''
Simple taskstats sample.
'''

import os

from pyroute2 import TaskStats

pid = os.getpid()
ts = TaskStats()
# bind is required in the case of generic netlink
ts.bind()
ret = ts.get_pid_stat(int(pid))[0]
# parsed structure
print(ret)
ts.close()



import errno

from pyroute2.common import basestring
from pyroute2.config import AF_BRIDGE
from pyroute2.netlink.rtnl.ndmsg import ndmsg
from pyroute2.requests.neighbour import NeighbourFieldFilter

from ..events import RescheduleException
from ..objects import AsyncObject


async def load_ndmsg(schema, sources, target, event):
    #
    # ignore events with ifindex == 0
    #
    if event['ifindex'] == 0:
        return
    #
    if event.get_attr('NDA_IFINDEX') is None:
        event['attrs'].append(('NDA_IFINDEX', event['ifindex']))
    #
    # AF_BRIDGE events
    #
    if event['family'] == AF_BRIDGE:
        #
        # bypass for now
        #
        try:
            await schema.load_netlink(
                'af_bridge_fdb', sources, target, event, propagate=True
            )
        except Exception:
            raise RescheduleException()

    else:
        await schema.load_netlink('neighbours', sources, target, event)


ndmsg_schema = (
    ndmsg.sql_schema()
    .unique_index('ifindex', 'NDA_DST', 'NDA_VLAN')
    .constraint('NDA_DST', "NOT NULL DEFAULT ''")
    .constraint('NDA_VLAN', "NOT NULL DEFAULT 0")
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_ifindex'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

brmsg_schema = (
    ndmsg.sql_schema()
    .unique_index('ifindex', 'flags', 'NDA_DST', 'NDA_LLADDR', 'NDA_VLAN')
    .constraint('NDA_LLADDR', "NOT NULL DEFAULT ''")
    .constraint('NDA_DST', "NOT NULL DEFAULT ''")
    .constraint('NDA_VLAN', "NOT NULL DEFAULT 0")
    .foreign_key(
        'interfaces',
        ('f_target', 'f_tflags', 'f_ifindex'),
        ('f_target', 'f_tflags', 'f_index'),
    )
)

init = {
    'specs': [['neighbours', ndmsg_schema], ['af_bridge_fdb', brmsg_schema]],
    'classes': [['neighbours', ndmsg], ['af_bridge_fdb', ndmsg]],
    'event_map': {ndmsg: [load_ndmsg]},
}


async def fallback_add(self, idx_req, req):
    async for msg in await self.sources[self['target']].api(self.api, 'dump'):
        await self.sources[self['target']].evq.put(msg)
    self.load_sql()


class Neighbour(AsyncObject):
    table = 'neighbours'
    msg_class = ndmsg
    field_filter = NeighbourFieldFilter
    api = 'neigh'

    @classmethod
    def _count(cls, view):
        if view.chain:
            return view.ndb.schema.fetchone(
                f'SELECT count(*) FROM {view.table} WHERE f_ifindex = ?',
                [view.chain['index']],
            )
        else:
            return view.ndb.schema.fetchone(
                f'SELECT count(*) FROM {view.table}'
            )

    @classmethod
    def _dump_where(cls, view):
        if view.chain:
            where = '''
                       WHERE
                           main.f_target = ? AND
                           main.f_ifindex = ?
                    '''
            values = [view.chain['target'], view.chain['index']]
        else:
            where = ''
            values = []
        return (where, values)

    @classmethod
    def summary(cls, view):
        req = f'''
              SELECT
                  main.f_target, main.f_tflags,
                  intf.f_IFLA_IFNAME, main.f_NDA_LLADDR, main.f_NDA_DST
              FROM
                  {cls.table} AS main
              INNER JOIN
                  interfaces AS intf
              ON
                  main.f_ifindex = intf.f_index
                  AND main.f_target = intf.f_target
              '''
        yield ('target', 'tflags', 'ifname', 'lladdr', 'dst')
        where, values = cls._dump_where(view)
        for record in view.ndb.schema.fetch(req + where, values):
            yield record

    def __init__(self, *argv, **kwarg):
        kwarg['iclass'] = ndmsg
        self.event_map = {ndmsg: "load_rtnlmsg"}
        super(Neighbour, self).__init__(*argv, **kwarg)
        self.fallback_for['add'][errno.EEXIST] = fallback_add

    def complete_key(self, key):
        if isinstance(key, dict):
            ret_key = key
        else:
            ret_key = {'target': self.ndb.localhost}

        if isinstance(key, basestring):
            ret_key['NDA_DST'] = key

        return super(Neighbour, self).complete_key(ret_key)

    def make_req(self, prime):
        req = super(Neighbour, self).make_req(prime)
        if 'vlan' in req and req['vlan'] == 0:
            req.pop('vlan')
        return req


class FDBRecord(Neighbour):
    table = 'af_bridge_fdb'
    msg_class = ndmsg
    api = 'fdb'

    @classmethod
    def summary(cls, view):
        for record in super(FDBRecord, cls).summary(view):
            yield record[:-1]

    def make_idx_req(self, prime):
        req = super(FDBRecord, self).make_req(prime)
        if 'NDA_VLAN' in req and req['NDA_VLAN'] == 0:
            req.pop('NDA_VLAN')
        return req



from dpkt.ip import IP

from pyroute2 import IPQSocket
from pyroute2.common import hexdump
from pyroute2.netlink.ipq import NF_ACCEPT

ip = IPQSocket()
ip.bind()
try:
    while True:
        msg = ip.get()[0]
        print("\n")
        print(hexdump(msg.raw))
        print(repr(IP(msg['payload'])))
        ip.verdict(msg['packet_id'], NF_ACCEPT)
except:
    pass
finally:
    ip.release()



from pyroute2.netlink import nla


class tuntap(nla):
    '''
    Fake data type
    '''

    prefix = 'IFTUN_'
    nla_map = (
        ('IFTUN_UNSPEC', 'none'),
        ('IFTUN_MODE', 'asciiz'),
        ('IFTUN_UID', 'uint32'),
        ('IFTUN_GID', 'uint32'),
        ('IFTUN_IFR', 'flags'),
    )

    class flags(nla):
        fields = (
            ('no_pi', 'B'),
            ('one_queue', 'B'),
            ('vnet_hdr', 'B'),
            ('tun_excl', 'B'),
            ('multi_queue', 'B'),
            ('persist', 'B'),
            ('nofilter', 'B'),
        )



from pyroute2.netlink.core import CoreSocket
from pyroute2.plan9 import Marshal9P


class Plan9Socket(CoreSocket):
    def __init__(self, use_socket=None):
        super().__init__(use_socket=use_socket)
        self.marshal = Marshal9P()
        self.spec['tag_field'] = 'tag'



'''
Monitor process exit
'''

from pyroute2 import TaskStats
from pyroute2.common import hexdump

pmask = ''

with open('/proc/cpuinfo', 'r') as f:
    for line in f.readlines():
        if line.startswith('processor'):
            pmask += ',' + line.split()[2]
pmask = pmask[1:]
ts = TaskStats()
ts.register_mask(pmask)
msg = ts.get()[0]
print(hexdump(msg.raw))
print(msg)

ts.deregister_mask(pmask)
ts.release()



import importlib
import logging
import os
import pkgutil
import struct
import sys
from socket import AF_INET, AF_INET6

from pyroute2 import config
from pyroute2.common import basestring, map_namespace
from pyroute2.config import AF_BRIDGE
from pyroute2.netlink import NLA_F_NESTED, nla, nlmsg, nlmsg_atoms
from pyroute2.netlink.rtnl.ifinfmsg.plugins import (
    bond,
    can,
    geneve,
    gtp,
    ipoib,
    ipvlan,
    team,
    tun,
    tuntap,
    vlan,
    vrf,
    vti,
    vti6,
    vxlan,
    xfrm,
)
from pyroute2.netlink.rtnl.iw_event import iw_event

log = logging.getLogger(__name__)

# it's simpler to double constants here, than to change all the
# module layout; but it is a subject of the future refactoring
RTM_NEWLINK = 16
RTM_DELLINK = 17
#


##
#
# tuntap flags
#
IFT_TUN = 0x0001
IFT_TAP = 0x0002
IFT_NO_PI = 0x1000
IFT_ONE_QUEUE = 0x2000
IFT_VNET_HDR = 0x4000
IFT_TUN_EXCL = 0x8000
IFT_MULTI_QUEUE = 0x0100
IFT_ATTACH_QUEUE = 0x0200
IFT_DETACH_QUEUE = 0x0400
# read-only
IFT_PERSIST = 0x0800
IFT_NOFILTER = 0x1000

##
#
# normal flags
#
IFF_UP = 0x1  # interface is up
IFF_BROADCAST = 0x2  # broadcast address valid
IFF_DEBUG = 0x4  # turn on debugging
IFF_LOOPBACK = 0x8  # is a loopback net
IFF_POINTOPOINT = 0x10  # interface is has p-p link
IFF_NOTRAILERS = 0x20  # avoid use of trailers
IFF_RUNNING = 0x40  # interface RFC2863 OPER_UP
IFF_NOARP = 0x80  # no ARP protocol
IFF_PROMISC = 0x100  # receive all packets
IFF_ALLMULTI = 0x200  # receive all multicast packets
IFF_MASTER = 0x400  # master of a load balancer
IFF_SLAVE = 0x800  # slave of a load balancer
IFF_MULTICAST = 0x1000  # Supports multicast
IFF_PORTSEL = 0x2000  # can set media type
IFF_AUTOMEDIA = 0x4000  # auto media select active
IFF_DYNAMIC = 0x8000  # dialup device with changing addresses
IFF_LOWER_UP = 0x10000  # driver signals L1 up
IFF_DORMANT = 0x20000  # driver signals dormant
IFF_ECHO = 0x40000  # echo sent packets

(IFF_NAMES, IFF_VALUES) = map_namespace('IFF', globals())

IFF_MASK = (
    IFF_UP
    | IFF_DEBUG
    | IFF_NOTRAILERS
    | IFF_NOARP
    | IFF_PROMISC
    | IFF_ALLMULTI
)

IFF_VOLATILE = (
    IFF_LOOPBACK
    | IFF_POINTOPOINT
    | IFF_BROADCAST
    | IFF_ECHO
    | IFF_MASTER
    | IFF_SLAVE
    | IFF_RUNNING
    | IFF_LOWER_UP
    | IFF_DORMANT
)

##
#
# gre flags
#
GRE_ACK = 0x0080
GRE_REC = 0x0700
GRE_STRICT = 0x0800
GRE_SEQ = 0x1000
GRE_KEY = 0x2000
GRE_ROUTING = 0x4000
GRE_CSUM = 0x8000

(GRE_NAMES, GRE_VALUES) = map_namespace('GRE_', globals())

##
#
# vlan filter flags
#
BRIDGE_VLAN_INFO_MASTER = 0x1  # operate on bridge device
BRIDGE_VLAN_INFO_PVID = 0x2  # ingress untagged
BRIDGE_VLAN_INFO_UNTAGGED = 0x4  # egress untagged
BRIDGE_VLAN_INFO_RANGE_BEGIN = 0x8  # range start
BRIDGE_VLAN_INFO_RANGE_END = 0x10  # range end
BRIDGE_VLAN_INFO_BRENTRY = 0x20  # global bridge vlan entry
(BRIDGE_VLAN_NAMES, BRIDGE_VLAN_VALUES) = map_namespace(
    'BRIDGE_VLAN_INFO', globals()
)

BRIDGE_VLAN_TUNNEL_UNSPEC = 0
BRIDGE_VLAN_TUNNEL_ID = 1
BRIDGE_VLAN_TUNNEL_VID = 2
BRIDGE_VLAN_TUNNEL_FLAGS = 3
BRIDGE_VLAN_TUNNEL_MAX = 4

BRIDGE_FLAGS_MASTER = 1
BRIDGE_FLAGS_SELF = 2
(BRIDGE_FLAGS_NAMES, BRIDGE_FLAGS_VALUES) = map_namespace(
    'BRIDGE_FLAGS', globals()
)

##
#
# XDP flags
#
XDP_FLAGS_UPDATE_IF_NOEXIST = 1 << 0
XDP_FLAGS_SKB_MODE = 1 << 1
XDP_FLAGS_DRV_MODE = 1 << 2
XDP_FLAGS_HW_MODE = 1 << 3
XDP_FLAGS_REPLACE = 1 << 4
XDP_FLAGS_MODES = XDP_FLAGS_SKB_MODE | XDP_FLAGS_DRV_MODE | XDP_FLAGS_HW_MODE
XDP_FLAGS_MASK = (
    XDP_FLAGS_UPDATE_IF_NOEXIST | XDP_FLAGS_MODES | XDP_FLAGS_REPLACE
)

(XDP_FLAGS_NAMES, XDP_FLAGS_VALUES) = map_namespace('XDP_FLAGS', globals())

states = (
    'UNKNOWN',
    'NOTPRESENT',
    'DOWN',
    'LOWERLAYERDOWN',
    'TESTING',
    'DORMANT',
    'UP',
)
state_by_name = {i[1]: i[0] for i in enumerate(states)}
state_by_code = dict(enumerate(states))
stats_names = (
    'rx_packets',
    'tx_packets',
    'rx_bytes',
    'tx_bytes',
    'rx_errors',
    'tx_errors',
    'rx_dropped',
    'tx_dropped',
    'multicast',
    'collisions',
    'rx_length_errors',
    'rx_over_errors',
    'rx_crc_errors',
    'rx_frame_errors',
    'rx_fifo_errors',
    'rx_missed_errors',
    'tx_aborted_errors',
    'tx_carrier_errors',
    'tx_fifo_errors',
    'tx_heartbeat_errors',
    'tx_window_errors',
    'rx_compressed',
    'tx_compressed',
)


def load_plugins_by_path(path):
    plugins = {}
    files = set(
        [
            x.split('.')[0]
            for x in filter(
                lambda x: x.endswith(('.py', '.pyc', '.pyo')), os.listdir(path)
            )
            if not x.startswith('_')
        ]
    )
    sys.path.append(path)
    for name in files:
        try:
            module = __import__(name, globals(), locals(), [], 0)
            register_kind = getattr(module, 'register_kind', name)
            plugins[register_kind] = getattr(module, register_kind)
        except:
            pass
    sys.path.pop()
    return plugins


def load_plugins_by_pkg(pkg):
    plugins = {}
    plugin_modules = {
        name: name.split('.')[-1]
        for loader, name, ispkg in pkgutil.iter_modules(
            path=pkg.__path__, prefix=pkg.__name__ + '.'
        )
    }

    # Hack to make it compatible with pyinstaller
    # plugin loading will work with and without pyinstaller
    # Inspired on:
    # https://github.com/webcomics/dosage/blob/master/dosagelib/loader.py
    # see: https://github.com/pyinstaller/pyinstaller/issues/1905
    importers = map(pkgutil.get_importer, pkg.__path__)
    toc = set()

    for importer in importers:
        if hasattr(importer, 'toc'):
            toc |= importer.toc

    for element in toc:
        if element.startswith(pkg.__name__) and element != pkg.__name__:
            plugin_modules[element] = element.split('.')[-1]

    for mod_path, mod_name in plugin_modules.items():
        if mod_name.startswith('_'):
            continue
        module = importlib.import_module(mod_path)
        register_kind = getattr(module, 'register_kind', mod_name)
        plugins[register_kind] = getattr(module, register_kind)

    return plugins


data_plugins = {}

for module in (
    bond,
    can,
    geneve,
    gtp,
    ipvlan,
    team,
    tuntap,
    tun,
    vlan,
    vrf,
    vti,
    vti6,
    vxlan,
    xfrm,
    ipoib,
):
    name = module.__name__.split('.')[-1]
    data_plugins[name] = getattr(module, name)

for pkg in config.data_plugins_pkgs:
    data_plugins.update(load_plugins_by_pkg(pkg))

for path in config.data_plugins_path:
    data_plugins.update(load_plugins_by_path(path))


class ifla_bridge_id(nla):
    fields = [('value', '=8s')]

    def encode(self):
        r_prio = struct.pack('H', self['prio'])
        r_addr = struct.pack(
            'BBBBBB', *[int(i, 16) for i in self['addr'].split(':')]
        )
        self['value'] = r_prio + r_addr
        nla.encode(self)

    def decode(self):
        nla.decode(self)
        r_prio = self['value'][:2]
        r_addr = self['value'][2:]
        self.value = {
            'prio': struct.unpack('H', r_prio)[0],
            'addr': ':'.join(
                '%02x' % (i) for i in struct.unpack('BBBBBB', r_addr)
            ),
        }


class protinfo_bridge(nla):
    prefix = 'IFLA_BRPORT_'
    nla_map = (
        ('IFLA_BRPORT_UNSPEC', 'none'),
        ('IFLA_BRPORT_STATE', 'uint8'),
        ('IFLA_BRPORT_PRIORITY', 'uint16'),
        ('IFLA_BRPORT_COST', 'uint32'),
        ('IFLA_BRPORT_MODE', 'uint8'),
        ('IFLA_BRPORT_GUARD', 'uint8'),
        ('IFLA_BRPORT_PROTECT', 'uint8'),
        ('IFLA_BRPORT_FAST_LEAVE', 'uint8'),
        ('IFLA_BRPORT_LEARNING', 'uint8'),
        ('IFLA_BRPORT_UNICAST_FLOOD', 'uint8'),
        ('IFLA_BRPORT_PROXYARP', 'uint8'),
        ('IFLA_BRPORT_LEARNING_SYNC', 'uint8'),
        ('IFLA_BRPORT_PROXYARP_WIFI', 'uint8'),
        ('IFLA_BRPORT_ROOT_ID', 'br_id'),
        ('IFLA_BRPORT_BRIDGE_ID', 'br_id'),
        ('IFLA_BRPORT_DESIGNATED_PORT', 'uint16'),
        ('IFLA_BRPORT_DESIGNATED_COST', 'uint16'),
        ('IFLA_BRPORT_ID', 'uint16'),
        ('IFLA_BRPORT_NO', 'uint16'),
        ('IFLA_BRPORT_TOPOLOGY_CHANGE_ACK', 'uint8'),
        ('IFLA_BRPORT_CONFIG_PENDING', 'uint8'),
        ('IFLA_BRPORT_MESSAGE_AGE_TIMER', 'uint64'),
        ('IFLA_BRPORT_FORWARD_DELAY_TIMER', 'uint64'),
        ('IFLA_BRPORT_HOLD_TIMER', 'uint64'),
        ('IFLA_BRPORT_FLUSH', 'flag'),
        ('IFLA_BRPORT_MULTICAST_ROUTER', 'uint8'),
        ('IFLA_BRPORT_PAD', 'uint64'),
        ('IFLA_BRPORT_MCAST_FLOOD', 'uint8'),
        ('IFLA_BRPORT_MCAST_TO_UCAST', 'uint8'),
        ('IFLA_BRPORT_VLAN_TUNNEL', 'uint8'),
        ('IFLA_BRPORT_BCAST_FLOOD', 'uint8'),
        ('IFLA_BRPORT_GROUP_FWD_MASK', 'uint16'),
        ('IFLA_BRPORT_NEIGH_SUPPRESS', 'uint8'),
        ('IFLA_BRPORT_ISOLATED', 'uint8'),
        ('IFLA_BRPORT_BACKUP_PORT', 'uint32'),
        ('IFLA_BRPORT_MRP_RING_OPEN', 'uint8'),
        ('IFLA_BRPORT_MRP_IN_OPEN', 'uint8'),
        ('IFLA_BRPORT_MCAST_EHT_HOSTS_LIMIT', 'uint32'),
        ('IFLA_BRPORT_MCAST_EHT_HOSTS_CNT', 'uint32'),
        ('IFLA_BRPORT_LOCKED', 'uint8'),
        ('IFLA_BRPORT_MAB', 'uint8'),
        ('IFLA_BRPORT_MCAST_N_GROUPS', 'uint32'),
        ('IFLA_BRPORT_MCAST_MAX_GROUPS', 'uint32'),
        ('IFLA_BRPORT_NEIGH_VLAN_SUPPRESS', 'uint8'),
        ('IFLA_BRPORT_BACKUP_NHID', 'uint32'),
    )

    class br_id(ifla_bridge_id):
        pass


class macvx_data(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_MACVLAN_UNSPEC', 'none'),
        ('IFLA_MACVLAN_MODE', 'mode'),
        ('IFLA_MACVLAN_FLAGS', 'flags'),
        ('IFLA_MACVLAN_MACADDR_MODE', 'macaddr_mode'),
        ('IFLA_MACVLAN_MACADDR', 'l2addr'),
        ('IFLA_MACVLAN_MACADDR_DATA', 'macaddr_data'),
        ('IFLA_MACVLAN_MACADDR_COUNT', 'uint32'),
    )

    class mode(nlmsg_atoms.uint32):
        value_map = {
            0: 'none',
            1: 'private',
            2: 'vepa',
            4: 'bridge',
            8: 'passthru',
            16: 'source',
        }

    class flags(nlmsg_atoms.uint16):
        value_map = {0: 'none', 1: 'nopromisc'}

    class macaddr_mode(nlmsg_atoms.uint32):
        value_map = {0: 'add', 1: 'del', 2: 'flush', 3: 'set'}

    class macaddr_data(nla):
        nla_map = ((4, 'IFLA_MACVLAN_MACADDR', 'l2addr'),)


class iptnl_data(nla):
    prefix = 'IFLA_'
    nla_map = (
        ('IFLA_IPIP_UNSPEC', 'none'),
        ('IFLA_IPIP_LINK', 'uint32'),
        ('IFLA_IPIP_LOCAL', 'ip4addr'),
        ('IFLA_IPIP_REMOTE', 'ip4addr'),
        ('IFLA_IPIP_TTL', 'uint8'),
        ('IFLA_IPIP_TOS', 'uint8'),
        ('IFLA_IPIP_ENCAP_LIMIT', 'uint8'),
        ('IFLA_IPIP_FLOWINFO', 'be32'),
        ('IFLA_IPIP_FLAGS', 'uint32'),
        ('IFLA_IPIP_PROTO', 'uint8'),
        ('IFLA_IPIP_PMTUDISC', 'uint8'),
        ('IFLA_IPIP_6RD_PREFIX', 'ip6addr'),
        ('IFLA_IPIP_6RD_RELAY_PREFIX', 'ip4addr'),
        ('IFLA_IPIP_6RD_PREFIXLEN', 'uint16'),
        ('IFLA_IPIP_6RD_RELAY_PREFIXLEN', 'uint16'),
        ('IFLA_IPIP_ENCAP_TYPE', 'uint16'),
        ('IFLA_IPIP_ENCAP_FLAGS', 'uint16'),
        ('IFLA_IPIP_ENCAP_SPORT', 'be16'),
        ('IFLA_IPIP_ENCAP_DPORT', 'be16'),
        ('IFLA_IPIP_COLLECT_METADATA', 'flag'),
        ('IFLA_IPIP_FWMARK', 'uint32'),
    )


class ifinfbase(object):
    '''
    Network interface message.

    C structure::

        struct ifinfomsg {
            unsigned char  ifi_family; /* AF_UNSPEC */
            unsigned short ifi_type;   /* Device type */
            int            ifi_index;  /* Interface index */
            unsigned int   ifi_flags;  /* Device flags  */
            unsigned int   ifi_change; /* change mask */
        };
    '''

    prefix = 'IFLA_'

    #
    # Changed from PRIMARY KEY to NOT NULL to support multiple
    # targets in one table, so we can collect info from multiple
    # systems.
    #
    # To provide data integrity one should use foreign keys,
    # but when you create a foreign key using interfaces as
    # the parent table, create also a unique index on
    # the fields specified in the foreign key definition.
    #
    # E.g.
    #
    # CREATE TABLE interfaces (f_target TEXT NOT NULL,
    #                          f_index INTEGER NOT NULL, ...)
    # CREATE TABLE routes (f_target TEXT NOT NULL,
    #                      f_RTA_OIF INTEGER, ...
    #                      FOREIGN KEY (f_target, f_RTA_OIF)
    #                      REFERENCES interfaces(f_target, f_index))
    # CREATE UNIQUE INDEX if_idx ON interfaces(f_target, f_index)
    #
    sql_constraints = {'index': 'NOT NULL'}
    sql_extra_fields = (('state', 'TEXT'),)
    lookup_fallbacks = {'index': 'ifname'}

    fields = (
        ('family', 'B'),
        ('__align', 'x'),
        ('ifi_type', 'H'),
        ('index', 'i'),
        ('flags', 'I'),
        ('change', 'I'),
    )

    nla_map = (
        ('IFLA_UNSPEC', 'none'),
        ('IFLA_ADDRESS', 'l2addr'),
        ('IFLA_BROADCAST', 'l2addr'),
        ('IFLA_IFNAME', 'asciiz'),
        ('IFLA_MTU', 'uint32'),
        ('IFLA_LINK', 'uint32'),
        ('IFLA_QDISC', 'asciiz'),
        ('IFLA_STATS', 'ifstats'),
        ('IFLA_COST', 'hex'),
        ('IFLA_PRIORITY', 'hex'),
        ('IFLA_MASTER', 'uint32'),
        ('IFLA_WIRELESS', 'wireless'),
        ('IFLA_PROTINFO', 'protinfo'),
        ('IFLA_TXQLEN', 'uint32'),
        ('IFLA_MAP', 'ifmap'),
        ('IFLA_WEIGHT', 'hex'),
        ('IFLA_OPERSTATE', 'state'),
        ('IFLA_LINKMODE', 'uint8'),
        ('IFLA_LINKINFO', 'ifinfo'),
        ('IFLA_NET_NS_PID', 'uint32'),
        ('IFLA_IFALIAS', 'asciiz'),
        ('IFLA_NUM_VF', 'uint32'),
        ('IFLA_VFINFO_LIST', 'vflist'),
        ('IFLA_STATS64', 'ifstats64'),
        ('IFLA_VF_PORTS', 'hex'),
        ('IFLA_PORT_SELF', 'hex'),
        ('IFLA_AF_SPEC', 'af_spec'),
        ('IFLA_GROUP', 'uint32'),
        ('IFLA_NET_NS_FD', 'netns_fd'),
        ('IFLA_EXT_MASK', 'uint32'),
        ('IFLA_PROMISCUITY', 'uint32'),
        ('IFLA_NUM_TX_QUEUES', 'uint32'),
        ('IFLA_NUM_RX_QUEUES', 'uint32'),
        ('IFLA_CARRIER', 'uint8'),
        ('IFLA_PHYS_PORT_ID', 'hex'),
        ('IFLA_CARRIER_CHANGES', 'uint32'),
        ('IFLA_PHYS_SWITCH_ID', 'hex'),
        ('IFLA_LINK_NETNSID', 'int32'),
        ('IFLA_PHYS_PORT_NAME', 'asciiz'),
        ('IFLA_PROTO_DOWN', 'uint8'),
        ('IFLA_GSO_MAX_SEGS', 'uint32'),
        ('IFLA_GSO_MAX_SIZE', 'uint32'),
        ('IFLA_PAD', 'hex'),
        ('IFLA_XDP', 'xdp'),
        ('IFLA_EVENT', 'uint32'),
        ('IFLA_NEW_NETNSID', 'be32'),
        ('IFLA_IF_NETNSID', 'uint32'),
        ('IFLA_CARRIER_UP_COUNT', 'uint32'),
        ('IFLA_CARRIER_DOWN_COUNT', 'uint32'),
        ('IFLA_NEW_IFINDEX', 'uint32'),
        ('IFLA_MIN_MTU', 'uint32'),
        ('IFLA_MAX_MTU', 'uint32'),
        ('IFLA_PROP_LIST', 'proplist'),
        ('IFLA_ALT_IFNAME', 'asciiz'),
        ('IFLA_PERM_ADDRESS', 'hex'),
        ('IFLA_PROTO_DOWN_REASON', 'down_reason'),
        ('IFLA_PARENT_DEV_NAME', 'asciiz'),
        ('IFLA_PARENT_DEV_BUS_NAME', 'asciiz'),
        ('IFLA_GRO_MAX_SIZE', 'uint32'),
        ('IFLA_TSO_MAX_SIZE', 'uint32'),
        ('IFLA_TSO_MAX_SEGS', 'uint32'),
    )

    @staticmethod
    def flags2names(flags, mask=0xFFFFFFFF):
        ret = []
        for flag in IFF_VALUES:
            if (flag & mask & flags) == flag:
                ret.append(IFF_VALUES[flag])
        return ret

    @staticmethod
    def names2flags(flags):
        ret = 0
        mask = 0
        for flag in flags:
            if flag[0] == '!':
                flag = flag[1:]
            else:
                ret |= IFF_NAMES[flag]
            mask |= IFF_NAMES[flag]
        return (ret, mask)

    def encode(self):
        # convert flags
        if isinstance(self['flags'], (set, tuple, list)):
            self['flags'], self['change'] = self.names2flags(self['flags'])
        return super(ifinfbase, self).encode()

    class down_reason(nla):
        nla_flags = NLA_F_NESTED
        prefix = 'IFLA_'
        nla_map = (
            ('IFLA_PROTO_DOWN_REASON_UNSPEC', 'none'),
            ('IFLA_PROTO_DOWN_REASON_MASK', 'uint32'),
            ('IFLA_PROTO_DOWN_REASON_VALUE', 'uint32'),
        )

    class netns_fd(nla):
        fields = [('value', 'I')]
        netns_run_dir = '/var/run/netns'
        netns_fd = None

        def encode(self):
            #
            # There are two ways to specify netns
            #
            # 1. provide fd to an open file
            # 2. provide a file name
            #
            # In the first case, the value is passed to the kernel
            # as is. In the second case, the object opens appropriate
            # file from `self.netns_run_dir` and closes it upon
            # `__del__(self)`
            if isinstance(self.value, int):
                self['value'] = self.value
            else:
                if isinstance(self.value, bytes):
                    self.value = self.value.decode('utf-8')
                if '/' in self.value:
                    netns_path = self.value
                else:
                    netns_path = '%s/%s' % (self.netns_run_dir, self.value)
                self.netns_fd = os.open(netns_path, os.O_RDONLY)
                self['value'] = self.netns_fd
                self.register_clean_cb(self.close)
            nla.encode(self)

        def close(self):
            if self.netns_fd is not None:
                os.close(self.netns_fd)

    class xdp(nla):
        nla_flags = NLA_F_NESTED
        prefix = 'IFLA_'
        nla_map = (
            ('IFLA_XDP_UNSPEC', 'none'),
            ('IFLA_XDP_FD', 'xdp_fd'),
            ('IFLA_XDP_ATTACHED', 'xdp_mode'),
            ('IFLA_XDP_FLAGS', 'xdp_flags'),
            ('IFLA_XDP_PROG_ID', 'uint32'),
            ('IFLA_XDP_DRV_PROG_ID', 'uint32'),
            ('IFLA_XDP_SKB_PROG_ID', 'uint32'),
            ('IFLA_XDP_HW_PROG_ID', 'uint32'),
            ('IFLA_XDP_EXPECTED_FD', 'xdp_fd'),
        )

        class xdp_fd(nlmsg_atoms.int32):
            sql_type = None

        class xdp_flags(nla):
            fields = [('value', 'I')]
            sql_type = 'INTEGER'

            def encode(self):
                v = self.value
                for flag in XDP_FLAGS_VALUES:
                    v &= ~flag

                if v != 0:
                    log.warning('possibly incorrect XDP flags')

                nla.encode(self)

        class xdp_mode(nlmsg_atoms.uint8):
            value_map = {
                0: None,
                1: 'xdp',
                2: 'xdpgeneric',
                3: 'xdpoffload',
                4: 'xdpmulti',
            }

    class proplist(nla):
        nla_flags = NLA_F_NESTED
        # Proplist has currently only IFLA_ALT_IFNAME, but start at same
        # index than IFLA_ALT_IFNAME in ifinfbase()
        nla_map = ((53, 'IFLA_ALT_IFNAME', 'asciiz'),)

        def altnames(self):
            return (
                attr[1]
                for attr in self["attrs"]
                if attr[0] == "IFLA_ALT_IFNAME"
            )

    class vflist(nla):
        nla_map = (('IFLA_VF_INFO_UNSPEC', 'none'), ('IFLA_VF_INFO', 'vfinfo'))

        class vfinfo(nla):
            prefix = 'IFLA_VF_'
            nla_map = (
                ('IFLA_VF_UNSPEC', 'none'),
                ('IFLA_VF_MAC', 'vf_mac'),
                ('IFLA_VF_VLAN', 'vf_vlan'),
                ('IFLA_VF_TX_RATE', 'vf_tx_rate'),
                ('IFLA_VF_SPOOFCHK', 'vf_spoofchk'),
                ('IFLA_VF_LINK_STATE', 'vf_link_state'),
                ('IFLA_VF_RATE', 'vf_rate'),
                ('IFLA_VF_RSS_QUERY_EN', 'vf_rss_query_en'),
                ('IFLA_VF_STATS', 'vf_stats'),
                ('IFLA_VF_TRUST', 'vf_trust'),
                ('IFLA_VF_IB_NODE_GUID', 'vf_ib_node_guid'),
                ('IFLA_VF_IB_PORT_GUID', 'vf_ib_port_guid'),
                ('IFLA_VF_VLAN_LIST', 'vf_vlist'),
            )

            class vf_ib_node_guid(nla):
                fields = (('vf', 'I'), ('ib_node_guid', '32B'))

                def decode(self):
                    nla.decode(self)
                    self['ib_node_guid'] = ':'.join(
                        ['%02x' % x for x in self['ib_node_guid'][4:12][::-1]]
                    )

                def encode(self):
                    encoded_guid = self['ib_node_guid'].split(':')[::-1]
                    self['ib_node_guid'] = (
                        [0] * 4 + [int(x, 16) for x in encoded_guid] + [0] * 20
                    )
                    nla.encode(self)

            class vf_ib_port_guid(nla):
                fields = (('vf', 'I'), ('ib_port_guid', '32B'))

                def decode(self):
                    nla.decode(self)
                    self['ib_port_guid'] = ':'.join(
                        ['%02x' % x for x in self['ib_port_guid'][4:12][::-1]]
                    )

                def encode(self):
                    encoded_guid = self['ib_port_guid'].split(':')[::-1]
                    self['ib_port_guid'] = (
                        [0] * 4 + [int(x, 16) for x in encoded_guid] + [0] * 20
                    )
                    nla.encode(self)

            class vf_mac(nla):
                fields = (('vf', 'I'), ('mac', '32B'))

                def decode(self):
                    nla.decode(self)
                    self['mac'] = ':'.join(
                        ['%02x' % x for x in self['mac'][:6]]
                    )

                def encode(self):
                    self['mac'] = [
                        int(x, 16) for x in self['mac'].split(':')
                    ] + [0] * 26
                    nla.encode(self)

            class vf_vlan(nla):
                fields = (('vf', 'I'), ('vlan', 'I'), ('qos', 'I'))

            class vf_tx_rate(nla):
                fields = (('vf', 'I'), ('tx_rate', 'I'))

            class vf_spoofchk(nla):
                fields = (('vf', 'I'), ('spoofchk', 'I'))

            class vf_link_state(nla):
                fields = (('vf', 'I'), ('link_state', 'I'))

            class vf_rate(nla):
                fields = (
                    ('vf', 'I'),
                    ('min_tx_rate', 'I'),
                    ('max_tx_rate', 'I'),
                )

            class vf_rss_query_en(nla):
                fields = (('vf', 'I'), ('rss_query_en', 'I'))

            class vf_stats(nla):
                nla_map = (
                    ('IFLA_VF_STATS_RX_PACKETS', 'uint64'),
                    ('IFLA_VF_STATS_TX_PACKETS', 'uint64'),
                    ('IFLA_VF_STATS_RX_BYTES', 'uint64'),
                    ('IFLA_VF_STATS_TX_BYTES', 'uint64'),
                    ('IFLA_VF_STATS_BROADCAST', 'uint64'),
                    ('IFLA_VF_STATS_MULTICAST', 'uint64'),
                    ('IFLA_VF_STATS_PAD', 'uint64'),
                    ('IFLA_VF_STATS_RX_DROPPED', 'uint64'),
                    ('IFLA_VF_STATS_TX_DROPPED', 'uint64'),
                )

            class vf_trust(nla):
                fields = (('vf', 'I'), ('trust', 'I'))

            class vf_vlist(nla):
                nla_map = (
                    ('IFLA_VF_VLAN_INFO_UNSPEC', 'none'),
                    ('IFLA_VF_VLAN_INFO', 'ivvi'),
                )

                class ivvi(nla):
                    fields = (
                        ('vf', 'I'),
                        ('vlan', 'I'),
                        ('qos', 'I'),
                        ('proto', '>H'),
                    )

    class wireless(iw_event):
        pass

    class state(nla):
        fields = (('value', 'B'),)
        sql_type = 'TEXT'

        def encode(self):
            self['value'] = state_by_name[self.value]
            nla.encode(self)

        def decode(self):
            nla.decode(self)
            self.value = state_by_code[self['value']]

    class ifstats(nla):
        fields = [(i, 'I') for i in stats_names]

    class ifstats64(nla):
        fields = [(i, 'Q') for i in stats_names]

    class ifmap(nla):
        fields = (
            ('mem_start', 'Q'),
            ('mem_end', 'Q'),
            ('base_addr', 'Q'),
            ('irq', 'H'),
            ('dma', 'B'),
            ('port', 'B'),
        )

    @staticmethod
    def protinfo(self, *argv, **kwarg):
        proto_map = {AF_BRIDGE: protinfo_bridge}
        return proto_map[self['family']]

    class ifinfo(nla):
        prefix = 'IFLA_INFO_'
        nla_map = (
            ('IFLA_INFO_UNSPEC', 'none'),
            ('IFLA_INFO_KIND', 'asciiz'),
            ('IFLA_INFO_DATA', 'info_data'),
            ('IFLA_INFO_XSTATS', 'hex'),
            ('IFLA_INFO_SLAVE_KIND', 'asciiz'),
            ('IFLA_INFO_SLAVE_DATA', 'info_slave_data'),
        )

        @staticmethod
        def info_slave_data(self, *argv, **kwarg):
            '''
            Return IFLA_INFO_SLAVE_DATA type based on
            IFLA_INFO_SLAVE_KIND or IFLA_INFO_KIND.
            '''
            kind = self.get_attr('IFLA_INFO_SLAVE_KIND')
            if kind is None:
                kind = self.get_attr('IFLA_INFO_KIND')
            data_map = {
                'bridge': self.bridge_slave_data,
                'bridge_slave': self.bridge_slave_data,
                'bond': self.bond_slave_data,
            }
            return data_map.get(kind, self.hex)

        class bridge_slave_data(protinfo_bridge):
            pass

        class bond_slave_data(nla):
            nla_map = (
                ('IFLA_BOND_SLAVE_UNSPEC', 'none'),
                ('IFLA_BOND_SLAVE_STATE', 'uint8'),
                ('IFLA_BOND_SLAVE_MII_STATUS', 'uint8'),
                ('IFLA_BOND_SLAVE_LINK_FAILURE_COUNT', 'uint32'),
                ('IFLA_BOND_SLAVE_PERM_HWADDR', 'l2addr'),
                ('IFLA_BOND_SLAVE_QUEUE_ID', 'uint16'),
                ('IFLA_BOND_SLAVE_AD_AGGREGATOR_ID', 'uint16'),
            )

        @staticmethod
        def info_data(self, *argv, **kwarg):
            '''
            The function returns appropriate IFLA_INFO_DATA
            type according to IFLA_INFO_KIND info. Return
            'hex' type for all unknown kind's and when the
            kind is not known.
            '''
            kind = self.get_attr('IFLA_INFO_KIND')
            return self.data_map.get(kind, self.hex)

        class veth_data(nla):
            nla_map = (
                ('VETH_INFO_UNSPEC', 'none'),
                ('VETH_INFO_PEER', 'info_peer'),
            )

            @staticmethod
            def info_peer(self, *argv, **kwarg):
                return ifinfveth

        class ipip_data(iptnl_data):
            pass

        class sit_data(iptnl_data):
            nla_map = [
                (x[0].replace('IPIP', 'SIT'), x[1]) for x in iptnl_data.nla_map
            ]

        class ip6tnl_data(nla):
            prefix = 'IFLA_'
            nla_map = (
                ('IFLA_IP6TNL_UNSPEC', 'none'),
                ('IFLA_IP6TNL_LINK', 'uint32'),
                ('IFLA_IP6TNL_LOCAL', 'ip6addr'),
                ('IFLA_IP6TNL_REMOTE', 'ip6addr'),
                ('IFLA_IP6TNL_TTL', 'uint8'),
                ('IFLA_IP6TNL_TOS', 'uint8'),
                ('IFLA_IP6TNL_ENCAP_LIMIT', 'uint8'),
                ('IFLA_IP6TNL_FLOWINFO', 'be32'),
                ('IFLA_IP6TNL_FLAGS', 'uint32'),
                ('IFLA_IP6TNL_PROTO', 'uint8'),
                ('IFLA_IP6TNL_PMTUDISC', 'uint8'),
                ('IFLA_IP6TNL_6RD_PREFIX', 'ip6addr'),
                ('IFLA_IP6TNL_6RD_RELAY_PREFIX', 'ip4addr'),
                ('IFLA_IP6TNL_6RD_PREFIXLEN', 'uint16'),
                ('IFLA_IP6TNL_6RD_RELAY_PREFIXLEN', 'uint16'),
                ('IFLA_IP6TNL_ENCAP_TYPE', 'uint16'),
                ('IFLA_IP6TNL_ENCAP_FLAGS', 'uint16'),
                ('IFLA_IP6TNL_ENCAP_SPORT', 'be16'),
                ('IFLA_IP6TNL_ENCAP_DPORT', 'be16'),
                ('IFLA_IP6TNL_COLLECT_METADATA', 'flag'),
                ('IFLA_IP6TNL_FWMARK', 'uint32'),
            )

        class gre_data(nla):
            prefix = 'IFLA_'
            nla_map = (
                ('IFLA_GRE_UNSPEC', 'none'),
                ('IFLA_GRE_LINK', 'uint32'),
                ('IFLA_GRE_IFLAGS', 'gre_flags'),
                ('IFLA_GRE_OFLAGS', 'gre_flags'),
                ('IFLA_GRE_IKEY', 'be32'),
                ('IFLA_GRE_OKEY', 'be32'),
                ('IFLA_GRE_LOCAL', 'ip4addr'),
                ('IFLA_GRE_REMOTE', 'ip4addr'),
                ('IFLA_GRE_TTL', 'uint8'),
                ('IFLA_GRE_TOS', 'uint8'),
                ('IFLA_GRE_PMTUDISC', 'uint8'),
                ('IFLA_GRE_ENCAP_LIMIT', 'uint8'),
                ('IFLA_GRE_FLOWINFO', 'be32'),
                ('IFLA_GRE_FLAGS', 'uint32'),
                ('IFLA_GRE_ENCAP_TYPE', 'uint16'),
                ('IFLA_GRE_ENCAP_FLAGS', 'uint16'),
                ('IFLA_GRE_ENCAP_SPORT', 'be16'),
                ('IFLA_GRE_ENCAP_DPORT', 'be16'),
                ('IFLA_GRE_COLLECT_METADATA', 'flag'),
                ('IFLA_GRE_IGNORE_DF', 'uint8'),
                ('IFLA_GRE_FWMARK', 'uint32'),
            )

            class gre_flags(nla):
                fields = [('value', '>H')]
                sql_type = 'INTEGER'

                def encode(self):
                    #
                    # for details see:
                    url = 'https://github.com/svinota/pyroute2/issues/531'

                    v = self.value
                    for flag in GRE_VALUES:
                        v &= ~flag
                    if v != 0:
                        log.warning(
                            'possibly incorrect GRE flags, ' 'see %s' % url
                        )
                    nla.encode(self)

        class ip6gre_data(nla):
            # Ostensibly the same as ip6gre_data except that local
            # and remote are ipv6 addrs.
            # As of Linux 4.8,IFLA_GRE_COLLECT_METADATA has not been
            # implemented for IPv6.
            # Linux uses the same enum names for v6 and v4 (in if_tunnel.h);
            # Here we name them IFLA_IP6GRE_xxx instead to avoid conflicts
            # with gre_data above.
            prefix = 'IFLA_'
            nla_map = (
                ('IFLA_IP6GRE_UNSPEC', 'none'),
                ('IFLA_IP6GRE_LINK', 'uint32'),
                ('IFLA_IP6GRE_IFLAGS', 'uint16'),
                ('IFLA_IP6GRE_OFLAGS', 'uint16'),
                ('IFLA_IP6GRE_IKEY', 'be32'),
                ('IFLA_IP6GRE_OKEY', 'be32'),
                ('IFLA_IP6GRE_LOCAL', 'ip6addr'),
                ('IFLA_IP6GRE_REMOTE', 'ip6addr'),
                ('IFLA_IP6GRE_TTL', 'uint8'),
                ('IFLA_IP6GRE_TOS', 'uint8'),
                ('IFLA_IP6GRE_PMTUDISC', 'uint8'),
                ('IFLA_IP6GRE_ENCAP_LIMIT', 'uint8'),
                ('IFLA_IP6GRE_FLOWINFO', 'be32'),
                ('IFLA_IP6GRE_FLAGS', 'uint32'),
                ('IFLA_IP6GRE_ENCAP_TYPE', 'uint16'),
                ('IFLA_IP6GRE_ENCAP_FLAGS', 'uint16'),
                ('IFLA_IP6GRE_ENCAP_SPORT', 'be16'),
                ('IFLA_IP6GRE_ENCAP_DPORT', 'be16'),
            )

        class macvlan_data(macvx_data):
            pass

        class macvtap_data(macvx_data):
            nla_map = [
                (x[0].replace('MACVLAN', 'MACVTAP'), x[1])
                for x in macvx_data.nla_map
            ]

        class bridge_data(nla):
            prefix = 'IFLA_'
            nla_map = (
                ('IFLA_BR_UNSPEC', 'none'),
                ('IFLA_BR_FORWARD_DELAY', 'uint32'),
                ('IFLA_BR_HELLO_TIME', 'uint32'),
                ('IFLA_BR_MAX_AGE', 'uint32'),
                ('IFLA_BR_AGEING_TIME', 'uint32'),
                ('IFLA_BR_STP_STATE', 'uint32'),
                ('IFLA_BR_PRIORITY', 'uint16'),
                ('IFLA_BR_VLAN_FILTERING', 'uint8'),
                ('IFLA_BR_VLAN_PROTOCOL', 'be16'),
                ('IFLA_BR_GROUP_FWD_MASK', 'uint16'),
                ('IFLA_BR_ROOT_ID', 'br_id'),
                ('IFLA_BR_BRIDGE_ID', 'br_id'),
                ('IFLA_BR_ROOT_PORT', 'uint16'),
                ('IFLA_BR_ROOT_PATH_COST', 'uint32'),
                ('IFLA_BR_TOPOLOGY_CHANGE', 'uint8'),
                ('IFLA_BR_TOPOLOGY_CHANGE_DETECTED', 'uint8'),
                ('IFLA_BR_HELLO_TIMER', 'uint64'),
                ('IFLA_BR_TCN_TIMER', 'uint64'),
                ('IFLA_BR_TOPOLOGY_CHANGE_TIMER', 'uint64'),
                ('IFLA_BR_GC_TIMER', 'uint64'),
                ('IFLA_BR_GROUP_ADDR', 'l2addr'),
                ('IFLA_BR_FDB_FLUSH', 'flag'),
                ('IFLA_BR_MCAST_ROUTER', 'uint8'),
                ('IFLA_BR_MCAST_SNOOPING', 'uint8'),
                ('IFLA_BR_MCAST_QUERY_USE_IFADDR', 'uint8'),
                ('IFLA_BR_MCAST_QUERIER', 'uint8'),
                ('IFLA_BR_MCAST_HASH_ELASTICITY', 'uint32'),
                ('IFLA_BR_MCAST_HASH_MAX', 'uint32'),
                ('IFLA_BR_MCAST_LAST_MEMBER_CNT', 'uint32'),
                ('IFLA_BR_MCAST_STARTUP_QUERY_CNT', 'uint32'),
                ('IFLA_BR_MCAST_LAST_MEMBER_INTVL', 'uint64'),
                ('IFLA_BR_MCAST_MEMBERSHIP_INTVL', 'uint64'),
                ('IFLA_BR_MCAST_QUERIER_INTVL', 'uint64'),
                ('IFLA_BR_MCAST_QUERY_INTVL', 'uint64'),
                ('IFLA_BR_MCAST_QUERY_RESPONSE_INTVL', 'uint64'),
                ('IFLA_BR_MCAST_STARTUP_QUERY_INTVL', 'uint64'),
                ('IFLA_BR_NF_CALL_IPTABLES', 'uint8'),
                ('IFLA_BR_NF_CALL_IP6TABLES', 'uint8'),
                ('IFLA_BR_NF_CALL_ARPTABLES', 'uint8'),
                ('IFLA_BR_VLAN_DEFAULT_PVID', 'uint16'),
                ('IFLA_BR_PAD', 'uint64'),
                ('IFLA_BR_VLAN_STATS_ENABLED', 'uint8'),
                ('IFLA_BR_MCAST_STATS_ENABLED', 'uint8'),
                ('IFLA_BR_MCAST_IGMP_VERSION', 'uint8'),
                ('IFLA_BR_MCAST_MLD_VERSION', 'uint8'),
            )

            class br_id(ifla_bridge_id):
                pass

        # IFLA_INFO_DATA plugin system prototype
        data_map = {
            'macvlan': macvlan_data,
            'macvtap': macvtap_data,
            'ipip': ipip_data,
            'sit': sit_data,
            'ip6tnl': ip6tnl_data,
            'gre': gre_data,
            'gretap': gre_data,
            'ip6gre': ip6gre_data,
            'ip6gretap': ip6gre_data,
            'veth': veth_data,
            'bridge': bridge_data,
            'bridge_slave': bridge_slave_data,
        }
        # expand supported interface types
        data_map.update(data_plugins)

        @classmethod
        def register_link_kind(cls, path=None, pkg=None, module=None):
            cls.data_map.update(data_plugins)
            if path is not None:
                cls.data_map.update(load_plugins_by_path(path))
            elif pkg is not None:
                cls.data_map.update(load_plugins_by_pkg(pkg))
            elif module is not None:
                for name, link_class in module.items():
                    cls.data_map[name] = link_class
            else:
                raise TypeError('path or pkg required')

        @classmethod
        def unregister_link_kind(cls, kind):
            return cls.data_map.pop(kind)

        @classmethod
        def list_link_kind(cls):
            return cls.data_map

    sql_extend = ((ifinfo, 'IFLA_LINKINFO'), (xdp, 'IFLA_XDP'))

    @staticmethod
    def af_spec(self, *argv, **kwarg):
        specs = {
            0: self.af_spec_inet,
            AF_INET: self.af_spec_inet,
            AF_INET6: self.af_spec_inet,
            AF_BRIDGE: self.af_spec_bridge,
        }
        return specs.get(self['family'], self.hex)

    class af_spec_bridge(nla):
        prefix = 'IFLA_BRIDGE_'
        # Bug-Url: https://github.com/svinota/pyroute2/issues/284
        # resolve conflict with link()/flags
        # IFLA_BRIDGE_FLAGS is for compatibility, in nla dicts
        # IFLA_BRIDGE_VLAN_FLAGS overrides it
        nla_map = (
            (0, 'IFLA_BRIDGE_FLAGS', 'uint16'),
            (0, 'IFLA_BRIDGE_VLAN_FLAGS', 'vlan_flags'),
            (1, 'IFLA_BRIDGE_MODE', 'uint16'),
            (2, 'IFLA_BRIDGE_VLAN_INFO', 'vlan_info'),
            (3, 'IFLA_BRIDGE_VLAN_TUNNEL_INFO', 'vlan_tunnel_info'),
        )

        class vlan_flags(nla):
            fields = [('value', 'H')]

            def encode(self):
                # convert flags
                if isinstance(self['value'], basestring):
                    self['value'] = BRIDGE_FLAGS_NAMES[
                        'BRIDGE_FLAGS_' + self['value'].upper()
                    ]
                nla.encode(self)

        class vlan_info(nla):
            prefix = ''
            fields = (('flags', 'H'), ('vid', 'H'))

            @staticmethod
            def flags2names(flags):
                ret = []
                for flag in BRIDGE_VLAN_VALUES:
                    if (flag & flags) == flag:
                        ret.append(BRIDGE_VLAN_VALUES[flag])
                return ret

            @staticmethod
            def names2flags(flags):
                ret = 0
                for flag in flags:
                    ret |= BRIDGE_VLAN_NAMES[
                        'BRIDGE_VLAN_INFO_' + flag.upper()
                    ]
                return ret

            def encode(self):
                # convert flags
                if isinstance(self['flags'], (set, tuple, list)):
                    self['flags'] = self.names2flags(self['flags'])
                return super(nla, self).encode()

        class vlan_tunnel_info(nla):
            prefix = 'IFLA_BRIDGE_VLAN_TUNNEL_'
            nla_map = (
                ('IFLA_BRIDGE_VLAN_TUNNEL_UNSPEC', 'none'),
                ('IFLA_BRIDGE_VLAN_TUNNEL_ID', 'uint32'),
                ('IFLA_BRIDGE_VLAN_TUNNEL_VID', 'uint16'),
                ('IFLA_BRIDGE_VLAN_TUNNEL_FLAGS', 'uint16'),
            )

    class af_spec_inet(nla):
        nla_map = (
            ('AF_UNSPEC', 'none'),
            ('AF_UNIX', 'hex'),
            ('AF_INET', 'inet'),
            ('AF_AX25', 'hex'),
            ('AF_IPX', 'hex'),
            ('AF_APPLETALK', 'hex'),
            ('AF_NETROM', 'hex'),
            ('AF_BRIDGE', 'hex'),
            ('AF_ATMPVC', 'hex'),
            ('AF_X25', 'hex'),
            ('AF_INET6', 'inet6'),
        )

        class inet(nla):
            #  ./include/linux/inetdevice.h: struct ipv4_devconf
            #  ./include/uapi/linux/ip.h
            field_names = (
                'dummy',
                'forwarding',
                'mc_forwarding',
                'proxy_arp',
                'accept_redirects',
                'secure_redirects',
                'send_redirects',
                'shared_media',
                'rp_filter',
                'accept_source_route',
                'bootp_relay',
                'log_martians',
                'tag',
                'arpfilter',
                'medium_id',
                'noxfrm',
                'nopolicy',
                'force_igmp_version',
                'arp_announce',
                'arp_ignore',
                'promote_secondaries',
                'arp_accept',
                'arp_notify',
                'accept_local',
                'src_vmark',
                'proxy_arp_pvlan',
                'route_localnet',
                'igmpv2_unsolicited_report_interval',
                'igmpv3_unsolicited_report_interval',
            )
            fields = [(i, 'I') for i in field_names]

        class inet6(nla):
            prefix = 'IFLA_'
            nla_map = (
                ('IFLA_INET6_UNSPEC', 'none'),
                ('IFLA_INET6_FLAGS', 'uint32'),
                ('IFLA_INET6_CONF', 'ipv6_devconf'),
                ('IFLA_INET6_STATS', 'ipv6_stats'),
                ('IFLA_INET6_MCAST', 'hex'),
                ('IFLA_INET6_CACHEINFO', 'ipv6_cache_info'),
                ('IFLA_INET6_ICMP6STATS', 'icmp6_stats'),
                ('IFLA_INET6_TOKEN', 'ip6addr'),
                ('IFLA_INET6_ADDR_GEN_MODE', 'uint8'),
            )

            class ipv6_devconf(nla):
                # ./include/uapi/linux/ipv6.h
                # DEVCONF_
                field_names = (
                    'forwarding',
                    'hop_limit',
                    'mtu',
                    'accept_ra',
                    'accept_redirects',
                    'autoconf',
                    'dad_transmits',
                    'router_solicitations',
                    'router_solicitation_interval',
                    'router_solicitation_delay',
                    'use_tempaddr',
                    'temp_valid_lft',
                    'temp_preferred_lft',
                    'regen_max_retry',
                    'max_desync_factor',
                    'max_addresses',
                    'force_mld_version',
                    'accept_ra_defrtr',
                    'accept_ra_pinfo',
                    'accept_ra_rtr_pref',
                    'router_probe_interval',
                    'accept_ra_rt_info_max_plen',
                    'proxy_ndp',
                    'optimistic_dad',
                    'accept_source_route',
                    'mc_forwarding',
                    'disable_ipv6',
                    'accept_dad',
                    'force_tllao',
                    'ndisc_notify',
                )
                fields = [(i, 'I') for i in field_names]

            class ipv6_cache_info(nla):
                # ./include/uapi/linux/if_link.h: struct ifla_cacheinfo
                fields = (
                    ('max_reasm_len', 'I'),
                    ('tstamp', 'I'),
                    ('reachable_time', 'I'),
                    ('retrans_time', 'I'),
                )

            class ipv6_stats(nla):
                # ./include/uapi/linux/snmp.h
                field_names = (
                    'num',
                    'inpkts',
                    'inoctets',
                    'indelivers',
                    'outforwdatagrams',
                    'outpkts',
                    'outoctets',
                    'inhdrerrors',
                    'intoobigerrors',
                    'innoroutes',
                    'inaddrerrors',
                    'inunknownprotos',
                    'intruncatedpkts',
                    'indiscards',
                    'outdiscards',
                    'outnoroutes',
                    'reasmtimeout',
                    'reasmreqds',
                    'reasmoks',
                    'reasmfails',
                    'fragoks',
                    'fragfails',
                    'fragcreates',
                    'inmcastpkts',
                    'outmcastpkts',
                    'inbcastpkts',
                    'outbcastpkts',
                    'inmcastoctets',
                    'outmcastoctets',
                    'inbcastoctets',
                    'outbcastoctets',
                    'csumerrors',
                    'noectpkts',
                    'ect1pkts',
                    'ect0pkts',
                    'cepkts',
                )
                fields = [(i, 'Q') for i in field_names]

            class icmp6_stats(nla):
                # ./include/uapi/linux/snmp.h
                field_names = (
                    'num',
                    'inmsgs',
                    'inerrors',
                    'outmsgs',
                    'outerrors',
                    'csumerrors',
                )
                fields = [(i, 'Q') for i in field_names]


class ifinfmsg(ifinfbase, nlmsg):
    def decode(self):
        nlmsg.decode(self)
        if self['flags'] & 1:
            self['state'] = 'up'
        else:
            self['state'] = 'down'


class ifinfveth(ifinfbase, nla):
    pass



from enum import IntEnum, IntFlag


class MessageType(IntEnum):
    BOOTREQUEST = 1  # Client to server
    BOOTREPLY = 2  # Server to client


class HardwareType(IntEnum):
    ETHERNET = 1  # Ethernet (10Mb)
    EXPERIMENTAL_ETHERNET = 2
    AMATEUR_RADIO = 3
    TOKEN_RING = 4
    FDDI = 8
    ATM = 19
    WIRELESS_IEEE_802_11 = 20


class Flag(IntFlag):
    UNICAST = 0x0000  # Unicast response requested
    BROADCAST = 0x8000  # Broadcast response requested



import builtins
import gc
import json
import logging
import multiprocessing
import os
import select
import signal
import socket
import struct
from collections import namedtuple
from typing import Any, Callable, Optional, Union

from pyroute2 import config
from pyroute2.common import USE_DEFAULT_TIMEOUT
from pyroute2.netlink import exceptions as pyroute2_exceptions

log = logging.getLogger(__name__)

ChildProcessReturnValue = namedtuple(
    'ChildProcessReturnValue', ('payload', 'fds')
)
ChildFuncReturnType = Union[None, ChildProcessReturnValue, bytearray, bytes]


def wrapper(
    ctrl: socket.socket,
    func: Callable[..., ChildFuncReturnType],
    argv: list[Any],
) -> None:
    '''Child function wrapper.

    This code will be executed in the child process after running fork().
    The internal data structures might be damaged, so the function `func`
    must be as simple as possible, and use only local variables of simple
    types.

    The garbage collector should be disabled as well to minimize possible
    deadlocks.

    If process doesn't response in time, it will get killed.
    '''
    gc.disable()
    payload: bytes = b''
    ret_data: bytes = b''
    fds: list[int] = []
    sockets: list[socket.socket] = []
    try:
        ret = func(*argv)
        if isinstance(ret, bytes):
            ret_data = ret
        if isinstance(ret, ChildProcessReturnValue):
            ret_data, sockets = ret
        if isinstance(ret_data, bytearray):
            ret_data = bytes(ret_data)
        if not isinstance(ret_data, bytes):
            raise TypeError('return values not supported')
        payload = struct.pack('B', 2) + ret_data
        if sockets:
            fds = [x.fileno() for x in sockets]
    except Exception as e:
        payload = struct.pack('B', 1) + json.dumps(
            {'exception': e.__class__.__name__, 'options': e.args}
        ).encode('utf-8')
        fds = []
    finally:
        socket.send_fds(ctrl, [payload], fds, len(fds))


class ChildProcess:
    ctrl_r: socket.socket
    ctrl_w: socket.socket

    def __init__(
        self, target: Callable[..., ChildFuncReturnType], args: list[Any]
    ):
        self._mode: str = config.child_process_mode
        self._target: Callable[..., ChildFuncReturnType] = target
        self._args: list[Any] = args
        self._proc: Optional[multiprocessing.Process] = None
        self._running: bool = False
        self._exitcode: Optional[int] = None
        self._pid: Optional[int] = None

    def close(self):
        self.stop()

    def __enter__(self):
        self.run()
        return self

    def __exit__(self, *_):
        self.close()

    @property
    def mode(self):
        return self._mode

    @property
    def pid(self):
        return self._pid

    def communicate(
        self, timeout: int = USE_DEFAULT_TIMEOUT
    ) -> tuple[bytes, list[int]]:
        '''Communicate with the child process.

        Raises:

        * struct.error -- error unpacking response from the child process
        * OSError -- OS level error communicating with the child process
        * TimeoutError -- the child process is alive, but doesn't response
        * RuntimeError -- the child process is dead or not started
        * TypeError -- error loading propagated exception
        '''
        if not self._running:
            raise RuntimeError('child process not started yet')
        if timeout == USE_DEFAULT_TIMEOUT:
            timeout = config.default_communicate_timeout
        rl, _, _ = select.select([self.ctrl_r], [], [], timeout)
        if not len(rl):
            # no data received within timeout
            # 1. the child process is killed
            # 2. the child process is stuck
            #
            # So first check the process, if it is killed, raise
            # RuntimeError, otherwise raise TimeoutError.
            #
            c_pid, w_status = os.waitpid(self.pid, os.WNOHANG)
            if c_pid != 0:
                # process is dead
                self._exitcode = os.waitstatus_to_exitcode(w_status)
                raise RuntimeError(
                    f'child process is dead, status: {self.exitcode}'
                )
            # process is alive, but stuck, kill it
            self.stop(kill=True, reason='no response from the child')
            raise TimeoutError(f'no response from the child pid {self.pid}')

        ret_data = b''
        # raises OSError
        (raw_data, fds, _, _) = socket.recv_fds(self.ctrl_r, 1024, 1)
        # get the return type
        # raises struct.error
        (ret_type,) = struct.unpack('B', raw_data[:1])
        raw_data = raw_data[1:]
        if ret_type == 1:
            # exception
            payload = json.loads(raw_data.decode('utf-8'))
            if not set(payload.keys()) == set(('exception', 'options')):
                raise TypeError('error loading child exception')
            if payload.get('exception') is not None:
                error_class = getattr(builtins, payload['exception'], None)
                if error_class is None:
                    error_class = getattr(
                        pyroute2_exceptions, payload['exception'], None
                    )
                if error_class is None:
                    error_class = Exception
                if not issubclass(error_class, Exception):
                    raise TypeError('error loading child error')
                raise error_class(*payload['options'])
        elif ret_type == 2:
            # raw_data
            ret_data = raw_data
        return ret_data, fds

    def get_data(self, timeout: int = USE_DEFAULT_TIMEOUT) -> bytes:
        return self.communicate(timeout)[0]

    def get_fds(self, timeout: int = USE_DEFAULT_TIMEOUT) -> list[int]:
        return self.communicate(timeout)[1]

    @property
    def proc(self) -> multiprocessing.Process:
        if self._proc is None:
            raise RuntimeError('not started')
        return self._proc

    @proc.setter
    def proc(self, value: Optional[multiprocessing.Process]) -> None:
        self._proc = value

    def _unsupported(self) -> Exception:
        return TypeError('unsupported mode')

    def run(self) -> None:
        if self._running:
            return
        self._running = True
        self.ctrl_r, self.ctrl_w = socket.socketpair(
            socket.AF_UNIX, socket.SOCK_DGRAM
        )
        if self.mode == 'fork':
            self._pid = os.fork()
            if self._pid == 0:
                wrapper(self.ctrl_w, self._target, self._args)
                os._exit(0)
        elif self.mode == 'mp':
            self.proc = multiprocessing.Process(
                target=wrapper, args=[self.ctrl_w, self._target, self._args]
            )
            self.proc.start()
            self._pid = self.proc.pid
        else:
            raise self._unsupported()

    def stop(self, kill: bool = False, reason: Optional[str] = None) -> None:
        if not self._running:
            return
        self._running = False
        self.ctrl_r.close()
        self.ctrl_w.close()
        if config.force_gc:
            gc.collect()
        if self.mode == 'fork':
            try:
                if kill:
                    os.kill(self.pid, signal.SIGKILL)
                else:
                    os.kill(self.pid, signal.SIGTERM)
                _, status = os.waitpid(self.pid, 0)
                self._exitcode = os.waitstatus_to_exitcode(status)
            except ProcessLookupError:
                # the child process has already exited
                pass
        elif self.mode == 'mp':
            if kill:
                self.proc.kill()
            else:
                self.proc.terminate()
            self.proc.join()
        else:
            raise self._unsupported()
        if reason is not None:
            log.warning(reason)

    @property
    def exitcode(self) -> Optional[int]:
        if self.mode == 'fork':
            return self._exitcode
        elif self.mode == 'mp':
            return self.proc.exitcode
        else:
            raise self._unsupported()



from socket import AF_INET, AF_INET6

import pytest
from common import Request, Result, run_test

from pyroute2.requests.address import AddressFieldFilter, AddressIPRouteFilter

config = {
    'filters': (
        {'class': AddressFieldFilter, 'argv': []},
        {'class': AddressIPRouteFilter, 'argv': ['add']},
    )
}


##
#
# IPv6 cacheinfo
#
@pytest.mark.parametrize(
    'spec,result',
    (
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'preferred_lft': 99,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'family': AF_INET6,
                    'cacheinfo': {
                        'ifa_preferred': 99,
                        'ifa_valid': 4294967295,
                    },
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'preferred': 99,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'family': AF_INET6,
                    'cacheinfo': {
                        'ifa_preferred': 99,
                        'ifa_valid': 4294967295,
                    },
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'valid_lft': 109,
                    'preferred': 99,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'family': AF_INET6,
                    'cacheinfo': {'ifa_preferred': 99, 'ifa_valid': 109},
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'valid': 109,
                    'preferred': 99,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'family': AF_INET6,
                    'cacheinfo': {'ifa_preferred': 99, 'ifa_valid': 109},
                }
            ),
        ),
    ),
    ids=('preferred_lft', 'preferred', 'valid_lft', 'valid'),
)
def test_cacheinfo(spec, result):
    return run_test(config, spec, result)


@pytest.mark.parametrize(
    'spec,result',
    (
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'valid': 109,
                    'preferred': 990,
                }
            ),
            Result({}),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '2001:db8::5678',
                    'prefixlen': 128,
                    'valid': 109,
                }
            ),
            Result({}),
        ),
    ),
    ids=('with-preferred', 'without-preferred'),
)
def test_cacheinfo_valid_fail(spec, result):
    with pytest.raises(ValueError) as e:
        run_test(config, spec, result)
    assert e.value.args == ('preferred_lft is greater than valid_lft',)


##
#
# broadcast tests: bool
#
@pytest.mark.parametrize(
    'spec,result',
    (
        (
            Request(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'prefixlen': 24,
                    'broadcast': True,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'prefixlen': 24,
                    'broadcast': '10.0.0.255',
                    'family': AF_INET,
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'prefixlen': 24,
                    'broadcast': False,
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'prefixlen': 24,
                    'family': AF_INET,
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'prefixlen': 28,
                    'broadcast': '10.0.0.15',
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'prefixlen': 28,
                    'broadcast': '10.0.0.15',
                    'family': AF_INET,
                }
            ),
        ),
    ),
    ids=('bool-true', 'bool-false', 'ipv4'),
)
def test_add_broadcast(spec, result):
    return run_test(config, spec, result)


##
#
# index format tests: int, list, tuple
#
result = Result(
    {
        'index': 1,
        'address': '10.0.0.1',
        'local': '10.0.0.1',
        'prefixlen': 24,
        'family': AF_INET,
    }
)


@pytest.mark.parametrize(
    'spec,result',
    (
        (
            Request({'index': 1, 'address': '10.0.0.1', 'prefixlen': 24}),
            result,
        ),
        (
            Request({'index': [1], 'address': '10.0.0.1', 'prefixlen': 24}),
            result,
        ),
        (
            Request({'index': (1,), 'address': '10.0.0.1', 'prefixlen': 24}),
            result,
        ),
    ),
    ids=('int', 'list', 'tuple'),
)
def test_index(spec, result):
    return run_test(config, spec, result)


@pytest.mark.parametrize(
    'spec,result',
    (
        (
            Request({'index': 1, 'address': '10.0.0.1'}),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'family': AF_INET,
                    'prefixlen': 32,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': '10.0.0.1', 'prefixlen': 16}),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'family': AF_INET,
                    'prefixlen': 16,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': '10.0.0.1/24'}),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'family': AF_INET,
                    'prefixlen': 24,
                }
            ),
        ),
        (
            Request(
                {'index': 1, 'address': '10.0.0.1', 'prefixlen': '255.0.0.0'}
            ),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'family': AF_INET,
                    'prefixlen': 8,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': '10.0.0.1/255.255.255.240'}),
            Result(
                {
                    'index': 1,
                    'address': '10.0.0.1',
                    'local': '10.0.0.1',
                    'family': AF_INET,
                    'prefixlen': 28,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': 'fc00::1'}),
            Result(
                {
                    'index': 1,
                    'address': 'fc00::1',
                    'family': AF_INET6,
                    'prefixlen': 128,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': 'fc00::1', 'prefixlen': 64}),
            Result(
                {
                    'index': 1,
                    'address': 'fc00::1',
                    'family': AF_INET6,
                    'prefixlen': 64,
                }
            ),
        ),
        (
            Request({'index': 1, 'address': 'fc00::1/48'}),
            Result(
                {
                    'index': 1,
                    'address': 'fc00::1',
                    'family': AF_INET6,
                    'prefixlen': 48,
                }
            ),
        ),
        (
            Request(
                {
                    'index': 1,
                    'address': 'fc00:0000:0000:0000:0000:0000:0000:0001/32',
                }
            ),
            Result(
                {
                    'index': 1,
                    'address': 'fc00::1',
                    'family': AF_INET6,
                    'prefixlen': 32,
                }
            ),
        ),
    ),
    ids=(
        'ipv4-default',
        'ipv4-prefixlen',
        'ipv4-split',
        'ipv4-prefixlen-dqn',
        'ipv4-split-dqn',
        'ipv6-default',
        'ipv6-prefixlen',
        'ipv6-split',
        'ipv6-compressed',
    ),
)
def test_family_and_prefix(spec, result):
    return run_test(config, spec, result)



'''
This tool is intended to decode existing data dumps produced with
other tools like tcpdump or strace, and print the data out in
JSON format.

The strace tool is not as convenient since version 4.13, as it
started to parse some of netlink messages at least partly,
rendering them useless for third party decoders. So if you plan
to use strace to obtain messages, be sure it is older than 4.13.
The strace related manual can be found in archive documentation
for older pyroute2 versions.

This manual is focused on pcap dumps.

An example session:

.. code-block:: console

    # set up netlink monitoring interface
    sudo ip link add dev nlmon0 type nlmon
    sudo ip link set dev nlmon0 up

    # dump the traffic into a pcap file
    # run netlink communication to be captured at the same time
    sudo tcpdump -i nlmon0 -w nl.pcap
    ^C

    # decode RTNL messages from the dump
    pyroute2-decoder \\
            -c pyroute2.netlink.rtnl.marshal.MarshalRtnl \\
            -d nl.pcap \\
            -m "ll_header{family=0}"

The result will be printed out in JSON format, so you can load
it directly from stdout, or use jq tool to navigate:

.. code-block:: console

    # print only pcap headers information
    pyroute2-decoder ... | jq '.[]."pcap header"'

pcap data dumps
~~~~~~~~~~~~~~~

This format is the default for `pyroute2-decoder`. To explicitly instruct
the decoder to use the pcap format, use `-f pcap` or `--format pcap`.

An ordinary everyday normal pcap dumps produced by tcpdump. The format
is described here shortly and only to the extent that is important for
the decoder. Please see other resources for detailed pcap format
descriptions. Pyroute2 decoder expect these headers in the pcap dump:

* Pcap file header. This header is being decoded, but not used by the
  tools as for now.
* Packet header. From this header the decoder uses only `incl_len` to
  properly read the stored data.
* Link layer header. From this header only the family field is used as
  for now, it can be matched with `ll_header{family=...}` expression.

hex data dumps
~~~~~~~~~~~~~~

Use `-f hex` or `--format hex`.

Just a raw data flow with no service headers added. The decoder uses
message headers to calculate the buffer lengths to read. This dump
can be obtained using strace or the IPBatch compiler.

Data should use hex bytes representation either in escaped or in
colon separated format. Equivalent variants:

* `\\\\x49\\\\x61\\\\x03\\\\x55`
* `49:61:03:55`

Comment strings start with `#`, comments and whitespaces are ignored.
A message example:

.. code-block::

    # ifinfmsg headers
    #
    # nlmsg header
    \\x84\\x00\\x00\\x00  # length
    \\x10\\x00          # type
    \\x05\\x06          # flags
    \\x49\\x61\\x03\\x55  # sequence number
    \\x00\\x00\\x00\\x00  # pid
    # RTNL header
    \\x00\\x00          # ifi_family
    \\x00\\x00          # ifi_type
    \\x00\\x00\\x00\\x00  # ifi_index
    \\x00\\x00\\x00\\x00  # ifi_flags
    \\x00\\x00\\x00\\x00  # ifi_change
    # ...


message classes
~~~~~~~~~~~~~~~

In order to properly debug the stream, one should specify either
a message class, or a marshal class:

.. code-block:: console

    # use a message class
    pyroute2-decoder \\
            -c pyroute2.netlink.generic.ipvs.ipvsmsg \\
            ...

    # use a marshal class
    pyroute2-decoder \\
            -c pyroute2.netlink.rtnl.marshal.MarshalRtnl \\
            ...

The decoder will try to use the specified class to decode every
matching message. That work well for generic protocols, but for other
protocols like RTNL it's more convenient to use marshal classes
that return corresponding message classes for different message
types.

generic protocols ids
~~~~~~~~~~~~~~~~~~~~~

Generic netlink protocols have dynamic IDs, so the first operation is to
get the ID. The message class used for that is `pyroute2.netlink.ctrlmsg`,
the request is `CTRL_CMD_GETFAMILY == 3`, and the response is
`CTRL_CMD_NEWFAMILY == 1`. The command is one byte right after the netlink
header, so the filters are:

* `ll_header{family=16}` match family 16, NETLINK_GENERIC
* `data{fmt='B', offset=16, value=1}` match one byte with
  value 1 by offset 16

Here is the code to get the family ID:

.. code-block:: console

    pyroute2-decoder \\
        -c pyroute2.netlink.ctrlmsg \\
        -d nl.pcap \\
        -m "ll_header{family=16} AND data{fmt='B', offset=16, value=1}" | \\
    jq \\
        '.[0].data.attrs[] | select(.[0] | contains("FAMILY"))'

    [
          "CTRL_ATTR_FAMILY_NAME",
          "IPVS"
    ]
    [
          "CTRL_ATTR_FAMILY_ID",
          37
    ]

Having the family ID you can filter out relevant messages. The filters:

* `ll_header{family=16}` match family 16, NETLINK_GENERIC
* `data{fmt='H', offset=4, value=37}` match IPVS family ID in
  the message header
* `data{fmt='B', offset=16, value=1}` match IPVS_CMD_NEW_SERVICE

.. code-block:: console

    pyroute2-decoder \\
        -c pyroute2.netlink.generic.ipvs.ipvsmsg \\
        -d nl0.pcap \\
        -m "ll_header{family=16} \\
            AND data{fmt='H', offset=4, value=37} \\
            AND data{fmt='B', offset=16, value=1}"


'''

import json

from pyroute2.common import hexdump
from pyroute2.decoder.args import parse_args
from pyroute2.decoder.loader import get_loader


def run():
    loader = get_loader(parse_args())
    ret = []
    for message in loader.data:
        ret.append(message.dump())
    print(json.dumps(ret, indent=4, default=lambda x: hexdump(x)))


if __name__ == "__main__":
    run()



'''
Utility to parse ifconfig, netstat etc.

PF_ROUTE may be effectively used only to get notifications. To fetch
info from the system we have to use ioctl or external utilities.

Maybe some day it will be ioctl. For now it's ifconfig and netstat.
'''

import re
import socket
import subprocess


class CMD(object):
    cmd = ['uname', '-s']

    def __init__(self, cmd=None):
        if cmd is not None:
            self.cmd = cmd

    def run(self):
        '''
        Run the command and get stdout
        '''
        stdout = stderr = ''
        try:
            process = subprocess.Popen(self.cmd, stdout=subprocess.PIPE)
            (stdout, stderr) = process.communicate()
        except Exception:
            process.kill()
        finally:
            process.wait()
        return stdout


class Route(CMD):
    cmd = ['netstat', '-rn']

    def parse(self, data):
        ret = []
        family = 0
        if isinstance(data, bytes):
            data = data.decode('utf-8')

        for line in data.split('\n'):
            if line == 'Internet:':
                family = socket.AF_INET
            elif line == 'Internet6:':
                # do NOT support IPv6 routes yet
                break

            sl = line.split()
            if len(sl) < 4:
                continue
            if sl[0] == 'Destination':
                # create the field map
                fmap = dict([(x[1], x[0]) for x in enumerate(sl)])
                if 'Netif' not in fmap:
                    fmap['Netif'] = fmap['Iface']
                continue

            route = {'family': family, 'attrs': []}

            #
            # RTA_DST
            dst = sl[fmap['Destination']]
            if dst != 'default':
                dst = dst.split('/')
                if len(dst) == 2:
                    dst, dst_len = dst
                else:
                    dst = dst[0]
                    if family == socket.AF_INET:
                        dst_len = 32
                    else:
                        dst_len = 128
                dst = dst.split('%')
                if len(dst) == 2:
                    dst, _ = dst
                else:
                    dst = dst[0]

                dst = '%s%s' % (dst, '.0' * (3 - dst.count('.')))
                route['dst_len'] = int(dst_len)
                route['attrs'].append(['RTA_DST', dst])
            #
            # RTA_GATEWAY
            gw = sl[fmap['Gateway']]
            if not gw.startswith('link') and not gw.find(':') >= 0:
                route['attrs'].append(['RTA_GATEWAY', sl[fmap['Gateway']]])
            #
            # RTA_OIF -- do not resolve it here! just save
            route['ifname'] = sl[fmap['Netif']]

            ret.append(route)
        return ret


class ARP(CMD):
    cmd = ['arp', '-an']

    def parse(self, data):
        ret = []
        f_dst = 1
        f_addr = 3
        f_ifname = 5
        if isinstance(data, bytes):
            data = data.decode('utf-8')

        for line in data.split('\n'):
            sl = line.split()
            if not sl:
                continue

            if sl[0] == 'Host':
                f_dst = 0
                f_addr = 1
                f_ifname = 2
                continue

            dst = sl[f_dst].strip('(').strip(')')
            addr = sl[f_addr].strip('(').strip(')')
            if addr == 'incomplete':
                continue

            ifname = sl[f_ifname]
            neighbour = {
                'ifindex': 0,
                'ifname': ifname,
                'family': 2,
                'attrs': [['NDA_DST', dst], ['NDA_LLADDR', addr]],
            }
            ret.append(neighbour)
        return ret


class Ifconfig(CMD):
    match = {'NR': re.compile(r'^\b').match}
    cmd = ['ifconfig', '-a']

    def parse_line(self, line):
        '''
        Dumb line parser:

        "key1 value1 key2 value2 something"
          -> {"key1": "value1", "key2": "value2"}
        '''
        ret = {}
        cursor = 0
        while cursor < (len(line) - 1):
            ret[line[cursor]] = line[cursor + 1]
            cursor += 2
        return ret

    def parse(self, data):
        '''
        Parse ifconfig output into netlink-compatible dicts::

            from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
            from pyroute2.bsd.util import Ifconfig

            def links()
                ifc = Ifconfig()
                data = ifc.run()
                for name, spec in ifc.parse(data)["links"].items():
                    yield ifinfmsg().load(spec)
        '''
        ifname = None
        kind = None
        ret = {'links': {}, 'addrs': {}}
        idx = 0
        info_data = {'attrs': None}

        if isinstance(data, bytes):
            data = data.decode('utf-8')

        for line in data.split('\n'):
            sl = line.split()
            pl = self.parse_line(sl)

            # type-specific
            if kind == 'gre' and 'inet' in sl and not info_data['attrs']:
                # first "inet" -- low-level addresses
                arrow = None
                try:
                    arrow = sl.index('->')
                except ValueError:
                    try:
                        arrow = sl.index('-->')
                    except ValueError:
                        continue
                if arrow is not None:
                    info_data['attrs'] = [
                        ('IFLA_GRE_LOCAL', sl[arrow - 1]),
                        ('IFLA_GRE_REMOTE', sl[arrow + 1]),
                    ]
                continue

            # first line -- ifname, flags, mtu
            if self.match['NR'](line):
                ifname = sl[0][:-1]
                kind = None
                idx += 1
                ret['links'][ifname] = link = {'index': idx, 'attrs': []}
                ret['addrs'][ifname] = addrs = []
                link['attrs'].append(['IFLA_IFNAME', ifname])
                #
                if ifname[:3] == 'gre':
                    kind = 'gre'
                    info_data = {'attrs': []}
                    linkinfo = {
                        'attrs': [
                            ('IFLA_INFO_KIND', kind),
                            ('IFLA_INFO_DATA', info_data),
                        ]
                    }
                    link['attrs'].append(['IFLA_LINKINFO', linkinfo])

                # extract flags
                try:
                    link['flags'] = int(sl[1].split('=')[1].split('<')[0])
                except Exception:
                    pass

                # extract MTU
                if 'mtu' in pl:
                    link['attrs'].append(['IFLA_MTU', int(pl['mtu'])])

            elif 'ether' in pl:
                link['attrs'].append(['IFLA_ADDRESS', pl['ether']])

            elif 'lladdr' in pl:
                link['attrs'].append(['IFLA_ADDRESS', pl['lladdr']])

            elif 'index' in pl:
                idx = int(pl['index'])
                link['index'] = int(pl['index'])

            elif 'inet' in pl:
                if ('netmask' not in pl) or ('inet' not in pl):
                    print(pl)
                    continue
                addr = {
                    'index': idx,
                    'family': socket.AF_INET,
                    'prefixlen': bin(int(pl['netmask'], 16)).count('1'),
                    'attrs': [['IFA_ADDRESS', pl['inet']]],
                }
                if 'broadcast' in pl:
                    addr['attrs'].append(['IFA_BROADCAST', pl['broadcast']])
                addrs.append(addr)
            elif 'inet6' in pl:
                if ('prefixlen' not in pl) or ('inet6' not in pl):
                    print(pl)
                    continue
                addr = {
                    'index': idx,
                    'family': socket.AF_INET6,
                    'prefixlen': int(pl['prefixlen']),
                    'attrs': [['IFA_ADDRESS', pl['inet6'].split('%')[0]]],
                }
                if 'scopeid' in pl:
                    addr['scope'] = int(pl['scopeid'], 16)
                addrs.append(addr)

        return ret



'''
..
    8<-----------------------------------------------------------
    Doctest setup.

    It is not rendered into html. For the docs, skip to tne
    CI test fixtures section below.

.. testsetup:: *

    from pyroute2.fixtures.doctest import *

..
    End of the doctest setup.
    8<-----------------------------------------------------------


CI test fixtures
----------------

Added in version 0.9.1.

The library provides a set of fixtures that can be used with pytest
to setup a simple test environment for functional network tests.

Usage:

.. testcode::

    # file: conftest.py
    pytest_plugins = [
        'pyroute2.fixtures.iproute',
        'pyroute2.fixtures.ndb',
        'pyroute2.fixtures.plan9'
    ]

    # file: my_test.py
    def test_my_code(sync_context):
        # here you have access to
        #
        sync_context.ipr        # IPRoute instance running in a netns
        sync_context.netns      # ready to use netns with a test link up
        sync_context.test_link  # test link in the netns
        sync_context.test_link.index    # interface index
        sync_context.test_link.ifname   # interface name
        sync_context.test_link.address  # MAC address

The fixtures set up a network namespace with a unique name, a dummy
interface within the namespace, and bring the interface up. They form
a tree of dependencies, so if you use e.g. `test_link_ifname` fixture,
you may be sure that the namespace and the interface are already set
up properly.

Fixtures dependencies diagram:

.. aafig::
    :scale: 80
    :textual:

                        +---------------------+
                        | `test_link`         |--+
                        +---------------------+  |
                             ^           |       v
    +---------------------+  |           |  +----------------------+
    | `test_link_index`   |__+           |  | `test_link_ifinfmsg` |
    |                     |  |           |  |                      |
    +---------------------+  |           |  +----------------------+
                             |           |       |
    +---------------------+  |           |       v
    | `test_link_address` |__+           |  +----------------------+
    |                     |  |           +->| netns                |
    +---------------------+  |           |  |                      |
                             |           |  +----------------------+
    +---------------------+  |           |
    | `test_link_ifname`  |__+           |
    |                     |  |           |
    +---------------------+  |           |
                             |           |
    +---------------------+  |           |
    | `async_context`     |__+           |
    |                     |_ | ___       |
    +---------------------+  |    |      |
                             |    |      |
    +---------------------+  |    |      |
    | `sync_context`      |__|    |      |
    |                     |_____  |      |
    +---------------------+     | |      |
                                | |      |
    +---------------------+     | |      |
    | `sync_ipr`          |<----+ |      |
    |                     |______ | _____+
    +---------------------+       |      |
                                  |      |
    +---------------------+       |      |
    | `async_ipr`         |<------+      |
    |                     |______________+
    +---------------------+              |
                                         |
    +---------------------+              |
    | `ndb`               |______________|
    |                     |
    +---------------------+

.. autofunction:: pyroute2.fixtures.iproute.nsname

.. autofunction:: pyroute2.fixtures.iproute.test_link_ifinfmsg

.. autofunction:: pyroute2.fixtures.iproute.test_link

.. autofunction:: pyroute2.fixtures.iproute.test_link_address

.. autofunction:: pyroute2.fixtures.iproute.test_link_index

.. autofunction:: pyroute2.fixtures.iproute.test_link_ifname

.. autofunction:: pyroute2.fixtures.iproute.async_ipr

.. autofunction:: pyroute2.fixtures.iproute.sync_ipr

.. autofunction:: pyroute2.fixtures.iproute.async_context

.. autofunction:: pyroute2.fixtures.iproute.sync_context

.. autofunction:: pyroute2.fixtures.ndb.ndb

.. autoclass:: pyroute2.fixtures.iproute.TestInterface
    :members:

.. autoclass:: pyroute2.fixtures.iproute.TestContext
    :members:

'''
